#  Copyright (c) 2021 zfit

import copy

import numpy as np
import pandas as pd
import pytest
import tensorflow as tf
import uproot

import zfit

obs1 = ('obs1', 'obs2', 'obs3')

example_data1 = np.random.random(size=(7, len(obs1)))


def create_data1():
    return zfit.Data.from_numpy(obs=obs1, array=example_data1)


@pytest.mark.parametrize("weights_factory", [lambda: None,
                                             lambda: 2. * tf.ones(shape=(1000,), dtype=tf.float64),
                                             lambda: np.random.normal(size=1000),
                                             lambda: 'eta1'])
def test_from_root(weights_factory):
    weights = weights_factory()

    from skhep_testdata import data_path

    path_root = data_path("uproot-Zmumu.root")

    branches = ['pt1', 'pt2', "phi2"]
    f = uproot.open(path_root)
    tree = f['events']

    true_data = tree.arrays(library='pd')

    data = zfit.Data.from_root(path=path_root, treepath='events', branches=branches, weights=weights)
    x = data.value()
    x_np = x.numpy()
    if weights is not None:
        weights_np = data.weights.numpy()
    else:
        weights_np = weights
    np.testing.assert_allclose(x_np, true_data[branches].values)
    if weights is not None:
        true_weights = weights if not isinstance(weights, str) else true_data[weights].values
        if isinstance(true_weights, tf.Tensor):
            true_weights = true_weights.numpy()
        np.testing.assert_allclose(weights_np, true_weights)
    else:
        assert weights_np is None


@pytest.mark.parametrize("weights_factory", [lambda: None,
                                             lambda: 2. * tf.ones(shape=(1000,), dtype=tf.float64),
                                             lambda: np.random.normal(size=1000), ])
def test_from_numpy(weights_factory):
    weights = weights_factory()

    example_data = np.random.random(size=(1000, len(obs1)))
    data = zfit.Data.from_numpy(obs=obs1, array=example_data, weights=weights)
    x = data.value()
    weights_from_data = data.weights
    if weights_from_data is not None:
        weights_from_data = weights_from_data.numpy()
    x_np = x
    np.testing.assert_array_equal(example_data, x_np)
    if isinstance(weights, tf.Tensor):
        weights = weights.numpy()
    if weights is not None:
        np.testing.assert_allclose(weights_from_data, weights)
    else:
        assert weights_from_data is None


def test_from_to_pandas():
    dtype = np.float32
    example_data_np = np.random.random(size=(1000, len(obs1)))
    example_data = pd.DataFrame(data=example_data_np, columns=obs1)
    data = zfit.Data.from_pandas(obs=obs1, df=example_data, dtype=dtype)
    x = data.value()
    assert x.dtype == dtype
    x_np = x.numpy()
    assert x_np.dtype == dtype
    np.testing.assert_array_equal(example_data_np.astype(dtype=dtype), x_np)

    # test auto obs retreavel
    example_data2 = pd.DataFrame(data=example_data_np, columns=obs1)
    data2 = zfit.Data.from_pandas(df=example_data2)
    assert data2.obs == obs1
    x2 = data2.value()
    x_np2 = x2.numpy()
    np.testing.assert_array_equal(example_data_np, x_np2)

    df = data2.to_pandas()
    assert all(df == example_data)


@pytest.mark.parametrize("weights_factory", [lambda: None,
                                             lambda: 2. * tf.ones(shape=(1000,), dtype=tf.float64),
                                             lambda: np.random.normal(size=1000), ])
def test_from_tensors(weights_factory):
    weights = weights_factory()
    true_tensor = 42. * tf.ones(shape=(1000, 1), dtype=tf.float64)
    data = zfit.Data.from_tensor(obs='obs1', tensor=true_tensor,
                                 weights=weights)

    weights_data = data.weights
    x = data.value()
    x_np = x.numpy()
    if isinstance(weights, tf.Tensor):
        weights = weights.numpy()

    np.testing.assert_allclose(x_np, true_tensor.numpy())
    if weights is not None:
        weights_data = weights_data.numpy()
        np.testing.assert_allclose(weights_data, weights)
    else:
        assert weights is None


def test_overloaded_operators():
    data1 = create_data1()
    a = data1 * 5.
    np.testing.assert_array_equal(5 * example_data1, a.numpy())
    np.testing.assert_array_equal(example_data1, data1.numpy())
    data_squared = data1 * data1
    np.testing.assert_allclose(example_data1 ** 2, data_squared.numpy(), rtol=1e-8)
    np.testing.assert_allclose(np.log(example_data1), tf.math.log(data1).numpy(), rtol=1e-8)


def test_sort_by_obs():
    data1 = create_data1()

    new_obs = (obs1[1], obs1[2], obs1[0])
    new_array = copy.deepcopy(example_data1)[:, np.array((1, 2, 0))]
    # new_array = np.array([new_array[:, 1], new_array[:, 2], new_array[:, 0]])
    assert data1.obs == obs1, "If this is not True, then the test will be flawed."
    with data1.sort_by_obs(new_obs):
        assert data1.obs == new_obs
        np.testing.assert_array_equal(new_array, data1.value().numpy())
        new_array2 = copy.deepcopy(new_array)[:, np.array((1, 2, 0))]
        # new_array2 = np.array([new_array2[:, 1], new_array2[:, 2], new_array2[:, 0]])
        new_obs2 = (new_obs[1], new_obs[2], new_obs[0])
        with data1.sort_by_obs(new_obs2):
            assert data1.obs == new_obs2
            np.testing.assert_array_equal(new_array2, data1.value().numpy())

        assert data1.obs == new_obs

    assert data1.obs == obs1
    np.testing.assert_array_equal(example_data1, data1.value().numpy())


def test_subdata():
    data1 = create_data1()
    new_obs = (obs1[0], obs1[1])
    new_array = copy.deepcopy(example_data1)[:, np.array((0, 1))]
    # new_array = np.array([new_array[:, 0], new_array])
    with data1.sort_by_obs(obs=new_obs):
        assert data1.obs == new_obs
        np.testing.assert_array_equal(new_array, data1.numpy())
        new_array2 = copy.deepcopy(new_array)[:, 1]
        # new_array2 = np.array([new_array2[:, 1]])
        new_obs2 = (new_obs[1],)
        with data1.sort_by_obs(new_obs2):
            assert data1.obs == new_obs2
            np.testing.assert_array_equal(new_array2, data1.value().numpy()[:, 0])

            with pytest.raises(ValueError):
                with data1.sort_by_obs(obs=new_obs):
                    data1.value().numpy()

    assert data1.obs == obs1
    np.testing.assert_array_equal(example_data1, data1.value())


@pytest.mark.parametrize("weights_factory", [lambda: None,
                                             lambda: np.random.normal(size=5), ])
def test_data_range(weights_factory):
    data1 = np.array([[1., 2],
                      [0, 1],
                      [-2, 1],
                      [-1, -1],
                      [-5, 10]])
    # data1 = data1.transpose()
    obs = ['obs1', 'obs2']
    lower1, lower2 = (0.5, 1), (-3, -2)
    upper1, upper2 = (1.5, 2.5), (-1.5, 1.5)
    space1 = zfit.Space(obs, limits=(lower1, upper1))
    space2 = zfit.Space(obs, limits=(lower2, upper2))
    data_range = space1 + space2
    cut_data1 = data1[np.array((0, 2)), :]

    weights = weights_factory()
    if weights is not None:
        cut_weights = weights[np.array((0, 2))]

    dataset = zfit.Data.from_tensor(obs=obs, tensor=data1, weights=weights)
    value_uncut = dataset.value()
    np.testing.assert_equal(data1, value_uncut.numpy())
    with dataset.set_data_range(data_range):
        value_cut = dataset.value()
        np.testing.assert_equal(cut_data1, value_cut.numpy())
        if dataset.has_weights:
            np.testing.assert_equal(cut_weights, dataset.weights.numpy())
        np.testing.assert_equal(data1, value_uncut.numpy())  # check  that the original did NOT change

    np.testing.assert_equal(cut_data1, value_cut.numpy())
    np.testing.assert_equal(data1, dataset.value().numpy())


def test_multidim_data_range():
    data1 = np.linspace((0, 5), (10, 15), num=11)
    data_true = np.linspace(10, 15, num=6)
    lower = ((5, 5),)
    upper = ((10, 15),)
    obs1 = "x"
    obs2 = "y"
    data_range = zfit.Space([obs1, obs2], limits=(lower, upper))
    dataset = zfit.Data.from_numpy(array=data1, obs=data_range)
    assert dataset.nevents.numpy() == 6
    with dataset.sort_by_obs(obs=obs1):
        assert dataset.nevents.numpy() == 6
    assert dataset.nevents.numpy() == 6
    with dataset.sort_by_obs(obs=obs2):
        assert dataset.nevents.numpy() == 6
        np.testing.assert_allclose(data_true, dataset.unstack_x().numpy())

    data2 = np.linspace((0, 5), (10, 15), num=11)[:, 1]
    data_range = zfit.Space([obs1], limits=(5, 15))
    dataset = zfit.Data.from_numpy(array=data2, obs=data_range)
    assert dataset.nevents.numpy() == 11



#  Copyright (c) 2021 zfit
from typing import Optional, Union

import tensorflow as tf
import tensorflow_addons as tfa
import tensorflow_probability as tfp

import zfit.z.numpy as znp

from .. import exception, z
from ..core.data import Data, sum_samples
from ..core.interfaces import ZfitPDF
from ..core.sample import accept_reject_sample
from ..core.space import supports
from ..util import ztyping
from ..util.exception import ShapeIncompatibleError, WorkInProgressError
from .functor import BaseFunctor


class FFTConvPDFV1(BaseFunctor):

    def __init__(self,
                 func: ZfitPDF,
                 kernel: ZfitPDF,
                 n: Optional[int] = None,
                 limits_func: Union[ztyping.LimitsType, float] = None,
                 limits_kernel: ztyping.LimitsType = None,
                 interpolation: Optional[str] = None,
                 obs: Optional[ztyping.ObsTypeInput] = None,
                 name: str = "FFTConvV1"):
        r"""*EXPERIMENTAL* Numerical Convolution pdf of `func` convoluted with `kernel` using FFT

        CURRENTLY ONLY 1 DIMENSIONAL!

        EXPERIMENTAL: Feedback is very welcome! Performance, which parameters to tune, which fail etc.

        TL;DR technical details:

          - FFT-like technique: discretization of function. Number of bins splits the kernel into `n` bins
            and uses the same binwidth for the func while extending it by the kernel space. Internally,
            `tf.nn.convolution` is used.
          - Then interpolation by either linear or spline function
          - The kernel is assumed to be "small enough" outside of it's `space` and points there won't be
            evaluated.

        The convolution of two (normalized) functions is defined as

        .. math::

            (f * g)(t) \triangleq\ \int_{-\infty}^\infty f(\tau) g(t - \tau) \, d\tau

        It defines the "smearing" of `func` by a `kernel`. This is when an element in `func` is
        randomly added to an element of `kernel`. While the sampling (the addition of elements) is rather
        simple to do computationally, the calculation of the convolutional PDF (if there is no analytic
        solution available) is not, as it requires:

            - an integral from -inf to inf
            - an integral *for every point of x that is requested*

        This can be solved with a few tricks. Instead of integrating to infinity, it is usually sufficient to
        integrate from a point where the function is "small enough".

        If the functions are arbitrary and with conditional dependencies,
        there is no way around an integral and another PDF has to be used. If the two functions are
        uncorrelated, a simplified version can be done by a discretization of the space (followed by a
        Fast Fourier Transfrom, after which the convolution becomes a simple multiplication) and a
        discrete convolution can be performed.

        An interpolation of the discrete convolution for the requested points `x` is performed afterwards.


        Args:
            func: PDF  with `pdf` method that takes x and returns the function value.
                Here x is a `Data` with the obs and limits of *limits*.
            kernel: PDF with `pdf` method that takes x acting as the kernel.
                Here x is a `Data` with the obs and limits of *limits*.
            n: Number of points *per dimension* to evaluate the kernel and pdf at.
                The higher the number of points, the more accurate the convolution at the cost
                of computing time. If `None`, a heuristic is used (default to 100 in 1 dimension).
            limits_func: Specify in which limits the `func` should
                be evaluated for the convolution:

                - If `None`, the limits from the `func` are used and extended by a
                 default value (relative 0.2).
                - If float: the fraction of the limit do be extended. 0 means no extension,
                  1 would extend the limits to each side by the same size resulting in
                  a tripled size (for 1 dimension).
                  As an example, the limits (1, 5) with a `limits_func` of 0.5 would result in
                  effective limits of (-1, 7), as 0.5 * (5 - 1) = 2 has been added to each side.
                - If a space with limits is used, this is taken as the range.

            limits_kernel: the limits of the kernel. Usually not needed to change and automatically
                taken from the kernel.
            interpolation: Specify the method that is used for interpolation. Available methods are:

                - 'linear': this is the default for any convolution > 1 dimensional. It is a
                  fast, linear interpolation between the evaluated points and approximates the
                  function reasonably well in case of high number of points and a smooth response.
                - 'spline' or `f'spline:{order}'`: a spline interpolation with polynomials. If
                  the order is not specified, a default is used. To specify the order, 'spline'
                  should be followed  an integer, separated by a colon as e.g. in 'spline:3'
                  to use a spline of order three.
                  This method is considerably more computationally expensive as it requires to solve
                  a system of equations. When using 1000+ points this can affect the runtime critical.
                  However, it provides better solution, a curve that is smooth even with less points
                  than for a linear interpolation.

            obs: Observables of the class. If not specified, automatically taken from `func`
            name: Human readable name of the PDF
        """
        from zfit import run
        run.assert_executing_eagerly()
        valid_interpolations = ('spline', 'linear')

        obs = func.space if obs is None else obs
        super().__init__(obs=obs, pdfs=[func, kernel], params={}, name=name)

        if self.n_obs > 1:
            raise WorkInProgressError("More than 1 dimensional convolutions are currently not supported."
                                      " If you need that, please open an issue on github.")

        if self.n_obs > 3:  # due to tf.nn.convolution not supporting more
            raise WorkInProgressError("More than 3 dimensional convolutions are currently not supported.")

        if interpolation is None:
            interpolation = 'spline' if self.n_obs == 1 else 'linear'

        spline_order = 3  # default
        if ':' in interpolation:
            interpolation, spline_order = interpolation.split(':')
            spline_order = int(spline_order)
        if interpolation not in valid_interpolations:
            raise ValueError(f"`interpolation` {interpolation} not known. Has to be one "
                             f"of the following: {valid_interpolations}")
        self._conv_interpolation = interpolation
        self._conv_spline_order = spline_order

        # get function limits
        if limits_func is None:
            limits_func = func.space
        limits_func = self._check_input_limits(limits=limits_func)
        if limits_func.n_limits == 0:
            raise exception.LimitsNotSpecifiedError("obs have to have limits to define where to integrate over.")
        if limits_func.n_limits > 1:
            raise WorkInProgressError("Multiple Limits not implemented")

        # get kernel limits
        if limits_kernel is None:
            limits_kernel = kernel.space
        limits_kernel = self._check_input_limits(limits=limits_kernel)

        if limits_kernel.n_limits == 0:
            raise exception.LimitsNotSpecifiedError("obs have to have limits to define where to integrate over.")
        if limits_kernel.n_limits > 1:
            raise WorkInProgressError("Multiple Limits not implemented")

        if func.n_obs != kernel.n_obs:
            raise ShapeIncompatibleError("Func and Kernel need to have (currently) the same number of obs,"
                                         f" currently are func: {func.n_obs} and kernel: {kernel.n_obs}")
        if not func.n_obs == limits_func.n_obs == limits_kernel.n_obs:
            raise ShapeIncompatibleError("Func and Kernel limits need to have (currently) the same number of obs,"
                                         f" are {limits_func.n_obs} and {limits_kernel.n_obs} with the func n_obs"
                                         f" {func.n_obs}")

        limits_func = limits_func.with_obs(obs)
        limits_kernel = limits_kernel.with_obs(obs)

        self._initargs = {
            'limits_kernel': limits_kernel,
            'limits_func': limits_func
        }

        if n is None:
            n = 51  # default kernel points
        if not n % 2:
            n += 1  # make it odd to have a unique shifting when using "same" in the convolution

        lower_func, upper_func = limits_func.rect_limits
        lower_kernel, upper_kernel = limits_kernel.rect_limits
        lower_sample = lower_func + lower_kernel
        upper_sample = upper_func + upper_kernel

        # TODO: what if kernel area is larger?
        if limits_kernel.rect_area() > limits_func.rect_area():
            raise WorkInProgressError("Currently, only kernels that are smaller than the func are supported."
                                      "Simply switch the two should resolve the problem.")

        # get finest resolution. Find the dimensions with the largest kernel-space to func-space ratio
        # We take the binwidth of the kernel as the overall binwidth and need to have the same binning in
        # the function as well
        area_ratios = (upper_sample - lower_sample) / (
                limits_kernel.rect_upper - limits_kernel.rect_lower)
        nbins_func_exact_max = znp.max(area_ratios * n)
        nbins_func = znp.ceil(nbins_func_exact_max)  # plus one and floor is like ceiling (we want more bins) with the
        # guarantee that we add one bin (e.g. if we hit exactly the boundaries, we add one.
        nbins_kernel = n
        # n = max(n, npoints_scaling)
        tf.assert_less(n - 1,  # so that for three dimension it's 999'999, not 10^6
                       tf.cast(1e6, tf.int32),
                       message="Number of points automatically calculated to be used for the FFT"
                               " based convolution exceeds 1e6. If you want to use this number - "
                               "or an even higher value - use explicitly the `n` argument.")

        binwidth = (upper_kernel - lower_kernel) / nbins_kernel
        to_extend = (binwidth * nbins_func - (
                upper_sample - lower_sample)) / 2  # how much we need to extend the func_limits
        # on each side in order to match the binwidth of the kernel
        lower_sample -= to_extend
        upper_sample += to_extend

        lower_valid = lower_sample + lower_kernel
        upper_valid = upper_sample + upper_kernel

        self._conv_limits = {'kernel': (lower_kernel, upper_kernel),
                             'valid': (lower_valid, upper_valid),
                             'func': (lower_sample, upper_sample),
                             'nbins_kernel': nbins_kernel,
                             'nbins_func': nbins_func}

    @z.function(wraps='model')
    def _unnormalized_pdf(self, x):

        lower_func, upper_func = self._conv_limits['func']
        nbins_func = self._conv_limits['nbins_func']
        x_funcs = tf.linspace(lower_func, upper_func, tf.cast(nbins_func, tf.int32))

        lower_kernel, upper_kernel = self._conv_limits['kernel']
        nbins_kernel = self._conv_limits["nbins_kernel"]
        x_kernels = tf.linspace(lower_kernel, upper_kernel, tf.cast(nbins_kernel, tf.int32))

        x_func = tf.meshgrid(*tf.unstack(x_funcs, axis=-1), indexing='ij')
        x_func = znp.transpose(x_func)
        x_func_flatish = znp.reshape(x_func, (-1, self.n_obs))
        data_func = Data.from_tensor(tensor=x_func_flatish, obs=self.obs)

        x_kernel = tf.meshgrid(*tf.unstack(x_kernels, axis=-1), indexing='ij')
        x_kernel = znp.transpose(x_kernel)
        x_kernel_flatish = znp.reshape(x_kernel, (-1, self.n_obs))
        data_kernel = Data.from_tensor(tensor=x_kernel_flatish, obs=self.obs)

        y_func = self.pdfs[0].pdf(data_func, norm_range=False)
        y_kernel = self.pdfs[1].pdf(data_kernel, norm_range=False)

        func_dims = [nbins_func] * self.n_obs
        kernel_dims = [nbins_kernel] * self.n_obs

        y_func = znp.reshape(y_func, func_dims)
        y_kernel = znp.reshape(y_kernel, kernel_dims)

        # flip the kernel to use the cross-correlation called `convolution function from TF
        # convolution = cross-correlation with flipped kernel
        # this introduces a shift and has to be corrected when interpolating/x_func
        # because the convolution has to be independent of the kernes **limits**
        # We assume they are symmetric when doing the FFT, so shift them back.
        y_kernel = tf.reverse(y_kernel, axis=range(self.n_obs))
        kernel_shift = (upper_kernel + lower_kernel) / 2
        x_func += kernel_shift
        lower_func += kernel_shift
        upper_func += kernel_shift

        # make rectangular grid
        y_func_rect = znp.reshape(y_func, func_dims)
        y_kernel_rect = znp.reshape(y_kernel, kernel_dims)

        # needed for multi dims?
        # if self.n_obs == 2:
        #     y_kernel_rect = tf.linalg.adjoint(y_kernel_rect)

        # get correct shape for tf.nn.convolution
        y_func_rect_conv = znp.reshape(y_func_rect, (1, *func_dims, 1))
        y_kernel_rect_conv = znp.reshape(y_kernel_rect, (*kernel_dims, 1, 1))

        conv = tf.nn.convolution(
            input=y_func_rect_conv,
            filters=y_kernel_rect_conv,
            strides=1,
            padding='SAME',
        )

        # needed for multidims?
        # if self.n_obs == 2:
        #     conv = tf.linalg.adjoint(conv[0, ..., 0])[None, ..., None]
        # conv = scipy.signal.convolve(
        #     y_func_rect,
        #     y_kernel_rect,
        #     mode='same'
        # )[None, ..., None]
        train_points = znp.expand_dims(x_func, axis=0)
        query_points = znp.expand_dims(x.value(), axis=0)
        if self.conv_interpolation == 'spline':
            conv_points = znp.reshape(conv, (1, -1, 1))
            prob = tfa.image.interpolate_spline(train_points=train_points,
                                                train_values=conv_points,
                                                query_points=query_points,
                                                order=self._conv_spline_order)
            prob = prob[0, ..., 0]
        elif self.conv_interpolation == 'linear':
            prob = tfp.math.batch_interp_regular_nd_grid(x=query_points[0],
                                                         x_ref_min=lower_func,
                                                         x_ref_max=upper_func,
                                                         y_ref=conv[0, ..., 0],
                                                         # y_ref=tf.reverse(conv[0, ..., 0], axis=[0]),
                                                         axis=-self.n_obs)
            prob = prob[0]

        return prob

    @property
    def conv_interpolation(self):
        return self._conv_interpolation

    @supports()
    def _sample(self, n, limits):
        # this is a custom implementation of sampling. Since the kernel and func are not correlated,
        # we can simply sample from both and add them. This is "trivial" compared to accept reject sampling
        # However, one large pitfall is that we cannot simply request n events from each pdf and then add them.
        # The kernel can move points out of the limits that we want, since it's "smearing" it.
        # E.g. with x sampled between limits, x + xkernel can be outside of limits.
        # Therefore we need to (repeatedly) sample from a) the func in the range of limits +
        # the kernel limits; to extend limits in the upper direction with the kernel upper limits and vice versa.
        # Therefore we (ab)use the accept reject sample: it samples until it is full. Everything has a constant
        # probability to be accepted.
        # This is maybe not the most efficient way to do and a more specialized (meaning taking care of less
        # special cases such as `accept_reject_sample` does) can be more efficient. However, sampling is not
        # supposed to be the bottleneck anyway.
        func = self.pdfs[0]
        kernel = self.pdfs[1]

        sample_and_weights = AddingSampleAndWeights(func=func,
                                                    kernel=kernel,
                                                    limits_func=self._initargs['limits_func'],
                                                    limits_kernel=self._initargs['limits_kernel'],
                                                    )

        return accept_reject_sample(
            lambda x: tf.ones(shape=tf.shape(x)[0], dtype=self.dtype),  # use inside?
            # all the points are inside
            n=n,
            limits=limits,
            sample_and_weights_factory=lambda: sample_and_weights,
            dtype=self.dtype,
            prob_max=1.,
            efficiency_estimation=0.9,
        )


class AddingSampleAndWeights:

    def __init__(self, func, kernel, limits_func, limits_kernel) -> None:
        super().__init__()
        self.func = func
        self.kernel = kernel
        self.limits_func = limits_func
        self.limits_kernel = limits_kernel

    def __call__(self, n_to_produce: Union[int, tf.Tensor], limits, dtype):
        kernel_lower, kernel_upper = self.kernel.space.rect_limits
        sample_lower, sample_upper = limits.rect_limits

        sample_ext_lower = sample_lower + kernel_lower
        sample_ext_upper = sample_upper + kernel_upper
        sample_ext_space = limits.with_limits((sample_ext_lower, sample_ext_upper))

        sample_func = self.func.sample(n=n_to_produce, limits=sample_ext_space)
        sample_kernel = self.kernel.sample(n=n_to_produce,
                                           limits=self.limits_kernel)  # no limits! it's the kernel, around 0
        sample = sum_samples(sample_func, sample_kernel, obs=limits, shuffle=True)
        sample = limits.filter(sample)
        n_drawn = tf.shape(sample)[0]
        from zfit import run
        if run.numeric_checks:
            tf.debugging.assert_positive(n_drawn,
                                         "Could not draw any samples. Check the limits of the func and kernel.")
        thresholds_unscaled = tf.ones(shape=(n_drawn,), dtype=sample.dtype)
        weights = thresholds_unscaled  # also "ones_like"
        weights_max = z.constant(1.)
        return sample, thresholds_unscaled * 0.5, weights, weights_max, n_drawn



"""A rich selection of analytically implemented Distributions (models) are available in `TensorFlow Probability.

<https://github.com/tensorflow/probability>`_. While their API is slightly different from the zfit models, it is similar
enough to be easily wrapped.

Therefore a convenient wrapper as well as a lot of implementations are provided.
"""
#  Copyright (c) 2021 zfit

from collections import OrderedDict

import tensorflow as tf
import tensorflow_probability as tfp
import tensorflow_probability.python.distributions as tfd

from zfit import z
from zfit.util.exception import (AnalyticIntegralNotImplemented,
                                 AnalyticSamplingNotImplemented)

from ..core.basepdf import BasePDF
from ..core.parameter import convert_to_parameter
from ..core.space import Space, supports
from ..settings import ztypes
from ..util import ztyping


# TODO: improve? while loop over `.sample`? Maybe as a fallback if not implemented?
@supports()
def tfd_analytic_sample(n: int, dist: tfd.Distribution, limits: ztyping.ObsTypeInput):
    """Sample analytically with a `tfd.Distribution` within the limits. No preprocessing.

    Args:
        n: Number of samples to get
        dist: Distribution to sample from
        limits: Limits to sample from within

    Returns:
        The sampled data with the number of samples and the number of observables.
    """
    lower_bound, upper_bound = limits.rect_limits
    lower_prob_lim = dist.cdf(lower_bound)
    upper_prob_lim = dist.cdf(upper_bound)

    shape = (n, 1)
    prob_sample = z.random.uniform(shape=shape, minval=lower_prob_lim,
                                   maxval=upper_prob_lim)
    prob_sample.set_shape((None, 1))
    try:
        sample = dist.quantile(prob_sample)
    except NotImplementedError:
        raise AnalyticSamplingNotImplemented
    sample.set_shape((None, limits.n_obs))
    return sample


class WrapDistribution(BasePDF):  # TODO: extend functionality of wrapper, like icdf
    """Baseclass to wrap tensorflow-probability distributions automatically."""

    def __init__(self, distribution, dist_params, obs, params=None, dist_kwargs=None, dtype=ztypes.float, name=None,
                 **kwargs):
        # Check if subclass of distribution?
        if dist_kwargs is None:
            dist_kwargs = {}

        if dist_params is None:
            dist_params = {}
        name = name or distribution.name
        if params is None:
            params = OrderedDict((k, p) for k, p in dist_params.items())
        else:
            params = OrderedDict((k, convert_to_parameter(p)) for k, p in params.items())

        super().__init__(obs=obs, dtype=dtype, name=name, params=params, **kwargs)

        self._distribution = distribution
        self.dist_params = dist_params
        self.dist_kwargs = dist_kwargs
        self._inverse_analytic_integral = []

    @property
    def distribution(self):

        params = self.dist_params
        if callable(params):
            params = params()
        kwargs = self.dist_kwargs
        if callable(kwargs):
            kwargs = kwargs()
        return self._distribution(**params, **kwargs, name=self.name + "_tfp")

    def _unnormalized_pdf(self, x: "zfit.Data", norm_range=False):
        value = z.unstack_x(x)  # TODO: use this? change shaping below?
        return self.distribution.prob(value=value, name="unnormalized_pdf")

    # TODO: register integral?
    @supports()
    def _analytic_integrate(self, limits, norm_range):
        lower, upper = limits._rect_limits_tf
        lower = z.unstack_x(lower)
        upper = z.unstack_x(upper)
        tf.debugging.assert_all_finite((lower, upper), "Are infinite limits needed? Causes troubles with NaNs")
        return self.distribution.cdf(upper) - self.distribution.cdf(lower)

    def _analytic_sample(self, n, limits: Space):
        return tfd_analytic_sample(n=n, dist=self.distribution, limits=limits)


# class KernelDensityTFP(WrapDistribution):
#
#     def __init__(self, loc: ztyping.ParamTypeInput, scale: ztyping.ParamTypeInput, obs: ztyping.ObsTypeInput,
#                  kernel: tfp.distributions.Distribution = tfp.distributions.Normal,
#                  weights: Union[None, np.ndarray, tf.Tensor] = None, name: str = "KernelDensity"):
#         """Kernel Density Estimation of loc and either a broadcasted or a per-loc scale with a Distribution as kernel.
#
#         Args:
#             loc: 1-D Tensor-like. The positions of the `kernel`. Determines how many kernels will be created.
#             scale: Broadcastable to the batch and event shape of the distribution. A scalar will simply broadcast
#                 to `loc` for a 1-D distribution.
#             obs: Observables
#             kernel: Distribution that is used as kernel
#             weights: Weights of each `loc`, can be None or Tensor-like with shape compatible with loc
#             name: Name of the PDF
#         """
#         if not isinstance(kernel,
#                           tfp.distributions.Distribution) and False:  # HACK remove False, why does test not work?
#             raise TypeError("Currently, only tfp distributions are supported as kernels. Please open an issue if this "
#                             "is too restrictive.")
#
#         if isinstance(loc, ZfitData):
#             if loc.weights is not None:
#                 if weights is not None:
#                     raise OverdefinedError("Cannot specify weights and use a `ZfitData` with weights.")
#                 else:
#                     weights = loc.weights
#
#         if weights is None:
#             weights = tf.ones_like(loc, dtype=tf.float64)
#         self._weights_loc = weights
#         self._weights_sum = z.reduce_sum(weights)
#         self._latent_loc = loc
#         params = {"scale": scale}
#         dist_params = {"loc": loc, "scale": scale}
#         super().__init__(distribution=kernel, dist_params=dist_params, obs=obs, params=params, dtype=ztypes.float,
#                          name=name)
#
#     def _unnormalized_pdf(self, x: "zfit.Data", norm_range=False):
#         value = znp.expand_dims(x.value(), -2)
#         new_shape = znp.concatenate([tf.shape(value)[:2], [tf.shape(self._latent_loc)[0], 4]], axis=0)
#         value = tf.broadcast_to(value, new_shape)
#         probs = self.distribution.prob(value=value, name="unnormalized_pdf")
#         # weights = znp.expand_dims(self._weights_loc, axis=-1)
#         weights = self._weights_loc
#         probs = z.reduce_sum(probs * weights, axis=-1) / self._weights_sum
#         return probs
#
#     @supports()
#     def _analytic_integrate(self, limits, norm_range):
#         lower, upper = limits.limits
#         if np.all(-np.array(lower) == np.array(upper)) and np.all(np.array(upper) == np.infty):
#             return z.reduce_sum(self._weights_loc)  # tfp distributions are normalized to 1
#         lower = z.to_real(lower[0], dtype=self.dtype)
#         # lower = tf.broadcast_to(lower, shape=(tf.shape(self._latent_loc)[0], limits.n_obs,))  # remove
#         upper = z.to_real(upper[0], dtype=self.dtype)
#         integral = self.distribution.cdf(upper) - self.distribution.cdf(lower)
#         integral = z.reduce_sum(integral * self._weights_loc, axis=-1) / self._weights_sum
#         return integral  # TODO: generalize for VectorSpaces


class Gauss(WrapDistribution):
    _N_OBS = 1

    def __init__(self, mu: ztyping.ParamTypeInput, sigma: ztyping.ParamTypeInput, obs: ztyping.ObsTypeInput,
                 name: str = "Gauss"):
        """Gaussian or Normal distribution with a mean (mu) and a standartdeviation (sigma).

        The gaussian shape is defined as

        .. math::
            f(x \\mid \\mu, \\sigma^2) = e^{ -\\frac{(x - \\mu)^{2}}{2\\sigma^2} }

        with the normalization over [-inf, inf] of

        .. math::
            \\frac{1}{\\sqrt{2\\pi\\sigma^2} }

        The normalization changes for different normalization ranges

        Args:
            mu: Mean of the gaussian dist
            sigma: Standard deviation or spread of the gaussian
            obs: Observables and normalization range the pdf is defined in
            name: Name of the pdf
        """
        mu, sigma = self._check_input_params(mu, sigma)
        params = OrderedDict((('mu', mu), ('sigma', sigma)))
        dist_params = lambda: dict(loc=mu.value(), scale=sigma.value())
        distribution = tfp.distributions.Normal
        super().__init__(distribution=distribution, dist_params=dist_params, obs=obs, params=params, name=name)


class ExponentialTFP(WrapDistribution):
    _N_OBS = 1

    def __init__(self, tau: ztyping.ParamTypeInput, obs: ztyping.ObsTypeInput, name: str = "Exponential"):
        (tau,) = self._check_input_params(tau)
        params = OrderedDict((('tau', tau),))
        dist_params = dict(rate=tau)
        distribution = tfp.distributions.Exponential
        super().__init__(distribution=distribution, dist_params=dist_params, obs=obs, params=params, name=name)


class Uniform(WrapDistribution):
    _N_OBS = 1

    def __init__(self, low: ztyping.ParamTypeInput, high: ztyping.ParamTypeInput, obs: ztyping.ObsTypeInput,
                 name: str = "Uniform"):
        """Uniform distribution which is constant between `low`, `high` and zero outside.

        Args:
            low: Below this value, the pdf is zero.
            high: Above this value, the pdf is zero.
            obs: Observables and normalization range the pdf is defined in
            name: Name of the pdf
        """
        low, high = self._check_input_params(low, high)
        params = OrderedDict((("low", low), ("high", high)))
        dist_params = lambda: dict(low=low.value(), high=high.value())
        distribution = tfp.distributions.Uniform
        super().__init__(distribution=distribution, dist_params=dist_params, obs=obs, params=params, name=name)


class TruncatedGauss(WrapDistribution):
    _N_OBS = 1

    def __init__(self, mu: ztyping.ParamTypeInput, sigma: ztyping.ParamTypeInput, low: ztyping.ParamTypeInput,
                 high: ztyping.ParamTypeInput, obs: ztyping.ObsTypeInput, name: str = "TruncatedGauss"):
        """Gaussian distribution that is 0 outside of `low`, `high`. Equivalent to the product of Gauss and Uniform.

        Args:
            mu: Mean of the gaussian dist
            sigma: Standard deviation or spread of the gaussian
            low: Below this value, the pdf is zero.
            high: Above this value, the pdf is zero.
            obs: Observables and normalization range the pdf is defined in
            name: Name of the pdf
        """
        mu, sigma, low, high = self._check_input_params(mu, sigma, low, high)
        params = OrderedDict((("mu", mu), ("sigma", sigma), ("low", low), ("high", high)))
        distribution = tfp.distributions.TruncatedNormal
        dist_params = lambda: dict(loc=mu.value(), scale=sigma.value(), low=low.value(), high=high.value())
        super().__init__(distribution=distribution, dist_params=dist_params,
                         obs=obs, params=params, name=name)


class Cauchy(WrapDistribution):
    _N_OBS = 1

    def __init__(self,
                 m: ztyping.ParamTypeInput,
                 gamma: ztyping.ParamTypeInput,
                 obs: ztyping.ObsTypeInput,
                 name: str = "Cauchy"):
        r"""Non-relativistic Breit-Wigner (Cauchy) PDF representing the energy distribution of a decaying particle.

        The (unnormalized) shape of the non-relativistic Breit-Wigner is given by

        .. math::

            \frac{1}{\gamma \left[1 + \left(\frac{x - m}{\gamma}\right)^2\right]}

        with :math:`m` the mean and :math:`\gamma` the width of the distribution.

        Args:
            m: Invariant mass of the unstable particle.
            gamma: Width of the shape.
            obs: Observables and normalization range the pdf is defined in
            name: Name of the PDF
        """
        m, gamma = self._check_input_params(m, gamma)
        params = OrderedDict((('m', m), ('gamma', gamma)))
        distribution = tfp.distributions.Cauchy
        dist_params = lambda: dict(loc=m.value(), scale=gamma.value())
        super().__init__(distribution=distribution, dist_params=dist_params,
                         obs=obs, params=params, name=name)


class Poisson(WrapDistribution):
    _N_OBS = 1

    def __init__(self,
                 lamb: ztyping.ParamTypeInput,
                 obs: ztyping.ObsTypeInput,
                 name: str = "Poisson"):
        """Poisson distribution, parametrized with an event rate parameter (lamb).

        The probability mass function of the Poisson distribution is given by

        .. math::
            f(x, \\lambda) = \\frac{\\lambda^{x}e^{-\\lambda}}{x!}

        Args:
            lamb: the event rate
            obs: Observables and normalization range the pdf is defined in
            name: Name of the PDF
        """
        (lamb,) = self._check_input_params(lamb)
        params = OrderedDict((('lamb', lamb),))
        dist_params = lambda: dict(rate=lamb.value())
        distribution = tfp.distributions.Poisson
        super().__init__(distribution=distribution, dist_params=dist_params,
                         obs=obs, params=params, name=name)



#  Copyright (c) 2021 zfit
import numpy as np
import pytest
from scipy.stats import crystalball

import zfit
from zfit.core.testing import tester
from zfit.models.physics import CrystalBall, DoubleCB

mu = -0.3
sigma = 1.1

alphal = 0.8
nl = 2.0

alphar = 1.4
nr = 4.0

bounds = (-6, 4)
lbounds = (bounds[0], mu)
rbounds = (mu, bounds[1])


def _cb_params_factory(name_add=""):
    mu_ = zfit.Parameter('mu_cb' + name_add, mu)
    sigma_ = zfit.Parameter('sigma_cb' + name_add, sigma)
    alphal_ = zfit.Parameter('alphal_cb' + name_add, alphal)
    nl_ = zfit.Parameter('nl_cb' + name_add, nl)
    return {"mu": mu_, "sigma": sigma_, "alpha": alphal_, "n": nl_}


tester.register_pdf(pdf_class=CrystalBall, params_factories=_cb_params_factory)


def sample_testing(pdf):
    sample = pdf.sample(n=1000, limits=(-0.5, 1.5))
    sample_np = sample.numpy()
    assert not any(np.isnan(sample_np))


def eval_testing(pdf, x):
    probs = pdf.pdf(x).numpy()
    assert not np.any(np.isnan(probs))
    return probs


def test_cb_integral():
    obs = zfit.Space('x', limits=bounds)

    mu_ = zfit.Parameter('mu_cb5', mu)
    sigma_ = zfit.Parameter('sigma_cb5', sigma)
    alphal_ = zfit.Parameter('alphal_cb5', alphal)
    nl_ = zfit.Parameter('nl_cb5', nl)

    cbl = CrystalBall(obs=obs, mu=mu_, sigma=sigma_, alpha=alphal_, n=nl_)
    int_limits = (-1, 3)
    integral_numeric = cbl.numeric_integrate(limits=int_limits, norm_range=False)

    integral = cbl.analytic_integrate(limits=int_limits, norm_range=False)
    integral_numeric = zfit.run(integral_numeric)
    integral = zfit.run(integral)

    assert pytest.approx(integral_numeric, integral, 1e-5)


def test_cb_dcb():
    obs = zfit.Space('x', limits=bounds)

    mu_ = zfit.Parameter('mu_cb5', mu)
    sigma_ = zfit.Parameter('sigma_cb5', sigma)
    alphal_ = zfit.Parameter('alphal_cb5', alphal)
    nl_ = zfit.Parameter('nl_cb5', nl)
    alphar_ = zfit.Parameter('alphar_cb5', alphar)
    nr_ = zfit.Parameter('nr_cb5', nr)

    cbl = CrystalBall(obs=obs, mu=mu_, sigma=sigma_, alpha=alphal_, n=nl_)
    cbr = CrystalBall(obs=obs, mu=mu_, sigma=sigma_, alpha=-alphar_, n=nr_)
    dcb = DoubleCB(obs=obs, mu=mu_, sigma=sigma_, alphal=alphal_, nl=nl_, alphar=alphar_, nr=nr_)

    sample_testing(cbl)
    sample_testing(cbr)
    sample_testing(dcb)

    x = np.random.normal(mu, sigma, size=10000)

    probsl = eval_testing(cbl, x)
    probsr = eval_testing(cbr, x)

    assert not any(np.isnan(probsl))
    assert not any(np.isnan(probsr))

    probsl_scipy = crystalball.pdf(x, beta=alphal, m=nl, loc=mu, scale=sigma)
    probsr_scipy = crystalball.pdf(-x + 2 * mu, beta=alphar, m=nr, loc=mu, scale=sigma)

    ratio_l = probsl_scipy / probsl
    ratio_r = probsr_scipy / probsr

    assert np.allclose(ratio_l, ratio_l[0])
    assert np.allclose(ratio_r, ratio_r[0])

    kwargs = dict(limits=(-5.0, mu), norm_range=lbounds)
    intl = cbl.integrate(**kwargs) - dcb.integrate(**kwargs)
    assert pytest.approx(intl.numpy(), abs=1e-3) == 0.
    intl = cbr.integrate(**kwargs) - dcb.integrate(**kwargs)
    assert pytest.approx(intl.numpy(), abs=1e-3) != 0

    kwargs = dict(limits=(mu, 2.0), norm_range=rbounds)
    intr = cbr.integrate(**kwargs) - dcb.integrate(**kwargs)
    assert pytest.approx(intr.numpy(), abs=1e-3) == 0.
    intr = cbl.integrate(**kwargs) - dcb.integrate(**kwargs)
    assert pytest.approx(intr.numpy(), abs=1e-3) != 0.

    xl = x[x <= mu]
    xr = x[x > mu]

    probs_dcb_l = eval_testing(dcb, xl)
    probs_dcb_r = eval_testing(dcb, xr)

    probsl_scipy = crystalball.pdf(xl, beta=alphal, m=nl, loc=mu, scale=sigma)
    probsr_scipy = crystalball.pdf(-xr + 2 * mu, beta=alphar, m=nr, loc=mu, scale=sigma)

    ratio_l = probsl_scipy / probs_dcb_l
    ratio_r = probsr_scipy / probs_dcb_r

    assert np.allclose(ratio_l, ratio_l[0])
    assert np.allclose(ratio_r, ratio_r[0])



#  Copyright (c) 2021 zfit

from typing import Callable, Iterable, Optional

import numdifftools
import tensorflow as tf

import zfit.z.numpy as znp

from ..util.container import convert_to_container
from ..util.deprecation import deprecated
from .tools import _auto_upcast
from .zextension import convert_to_tensor


def poly_complex(*args, real_x=False):
    """Complex polynomial with the last arg being x.

    Args:
        *args: Coefficients of the polynomial
        real_x: If True, x is assumed to be real.

    Returns:
    """
    from .. import z

    args = list(args)
    x = args.pop()
    if real_x is not None:
        pow_func = znp.power
    else:
        pow_func = z.nth_pow
    return tf.add_n([coef * z.to_complex(pow_func(x, p)) for p, coef in enumerate(args)])


def numerical_gradient(func: Callable, params: Iterable["zfit.Parameter"]) -> tf.Tensor:
    """Calculate numerically the gradient of func() with respect to `params`.

    Args:
        func: Function without arguments that depends on `params`
        params: Parameters that `func` implicitly depends on and with respect to which the
            derivatives will be taken.

    Returns:
        Gradients
    """
    from ..core.parameter import assign_values

    params = convert_to_container(params)

    def wrapped_func(param_values):
        assign_values(params, param_values)
        value = func()
        if hasattr(value, 'numpy'):
            value = value.numpy()
        return value

    param_vals = znp.stack(params)
    original_vals = [param.value() for param in params]
    grad_func = numdifftools.Gradient(wrapped_func, order=2, base_step=1e-4)
    if tf.executing_eagerly():
        grad_vals = grad_func(param_vals)
        gradient = convert_to_tensor(grad_vals)
    else:
        gradient = tf.numpy_function(grad_func, inp=[param_vals],
                                     Tout=tf.float64)
    if gradient.shape == ():
        gradient = znp.reshape(gradient, newshape=(1,))
    gradient.set_shape(param_vals.shape)
    assign_values(params, original_vals)
    return gradient


def numerical_value_gradient(func: Callable, params: Iterable["zfit.Parameter"]) -> [tf.Tensor, tf.Tensor]:
    """Calculate numerically the gradients of `func()` with respect to `params`, also returns the value of `func()`.

    Args:
        func: Function without arguments that depends on `params`
        params: Parameters that `func` implicitly depends on and with respect to which the
            derivatives will be taken.

    Returns:
        Value, gradient
    """
    return func(), numerical_gradient(func, params)


deprecated(None, "Use `numerical_value_gradient` instead.")


def numerical_value_gradients(*args, **kwargs):
    return numerical_value_gradients(*args, **kwargs)


def numerical_hessian(func: Optional[Callable],
                      params: Iterable["zfit.Parameter"],
                      hessian=None) -> tf.Tensor:
    """Calculate numerically the hessian matrix of func with respect to `params`.

    Args:
        func: Function without arguments that depends on `params`
        params: Parameters that `func` implicitly depends on and with respect to which the
            derivatives will be taken.

    Returns:
        Hessian matrix
    """
    from ..core.parameter import assign_values

    params = convert_to_container(params)

    def wrapped_func(param_values):
        assign_values(params, param_values)
        value = func()
        if hasattr(value, 'numpy'):
            value = value.numpy()
        return value

    param_vals = znp.stack(params)
    original_vals = [param.value() for param in params]

    if hessian == 'diag':
        hesse_func = numdifftools.Hessdiag(wrapped_func,
                                           order=2,
                                           # TODO: maybe add step to remove numerical problems?
                                           base_step=1e-4
                                           )
    else:
        hesse_func = numdifftools.Hessian(wrapped_func,
                                          order=2,
                                          base_step=1e-4,
                                          )
    if tf.executing_eagerly():
        computed_hessian = convert_to_tensor(hesse_func(param_vals))
    else:
        computed_hessian = tf.numpy_function(hesse_func, inp=[param_vals],
                                             Tout=tf.float64)
    n_params = param_vals.shape[0]
    if hessian == 'diag':
        computed_hessian.set_shape((n_params,))
    else:
        computed_hessian.set_shape((n_params, n_params))

    assign_values(params, original_vals)
    return computed_hessian


def numerical_value_gradient_hessian(func: Optional[Callable],
                                     params: Iterable["zfit.Parameter"],
                                     gradient: Optional[Callable] = None,
                                     hessian: Optional[str] = None) -> [tf.Tensor, tf.Tensor, tf.Tensor]:
    """Calculate numerically the gradients and hessian matrix of `func()` wrt `params`; also return `func()`.

    Args:
        func: Function without arguments that depends on `params`
        params: Parameters that `func` implicitly depends on and with respect to which the
            derivatives will be taken.

    Returns:
        Value, gradient and hessian matrix
    """
    if params is None:
        raise ValueError("params cannot be None")
    if func is None and gradient is None:
        raise ValueError("Either func or grad has to be given")
    value, gradients = numerical_value_gradient(func, params)
    hessian = numerical_hessian(func, params, hessian=hessian)

    return value, gradients, hessian


@deprecated(None, "Use `numerical_value_gradient_hessian` instead.")
def numerical_value_gradients_hessian(*args, **kwargs):
    return numerical_value_gradient_hessian(*args, **kwargs)


def autodiff_gradient(func: Callable, params: Iterable["zfit.Parameter"]) -> tf.Tensor:
    """Calculate using autodiff the gradients of `func()` wrt `params`.

    Automatic differentiation (autodiff) is a way of retreiving the derivative of x wrt y. It works by consecutively
    applying the chain rule. All that is needed is that every operation knows its own derivative.
    TensorFlow implements this and anything using `tf.*` operations only can use this technique.

        Args:
            func: Function without arguments that depends on `params`
            params: Parameters that `func` implicitly depends on and with respect to which the
                derivatives will be taken.

        Returns:
            Gradient
    """
    return autodiff_value_gradients(func, params)[1]


def autodiff_value_gradient(func: Callable, params: Iterable["zfit.Parameter"]) -> [tf.Tensor, tf.Tensor]:
    """Calculate using autodiff the gradients of `func()` wrt `params`; also return `func()`.

    Automatic differentiation (autodiff) is a way of retreiving the derivative of x wrt y. It works by consecutively
    applying the chain rule. All that is needed is that every operation knows its own derivative.
    TensorFlow implements this and anything using `tf.*` operations only can use this technique.

        Args:
            func: Function without arguments that depends on `params`
            params: Parameters that `func` implicitly depends on and with respect to which the
                derivatives will be taken.

        Returns:
            Value and gradient
    """
    with tf.GradientTape(persistent=False,  # needs to be persistent for a call from hessian.
                         watch_accessed_variables=False) as tape:
        tape.watch(params)
        value = func()
    gradients = tape.gradient(value, sources=params)
    return value, gradients


@deprecated(None, "Use `autodiff_value_gradient` instead.")
def autodiff_value_gradients(*args, **kwargs):
    return autodiff_value_gradient(*args, **kwargs)


def autodiff_hessian(func: Callable, params: Iterable["zfit.Parameter"], hessian=None) -> tf.Tensor:
    """Calculate using autodiff the hessian matrix of `func()` wrt `params`.

    Automatic differentiation (autodiff) is a way of retrieving the derivative of x wrt y. It works by consecutively
    applying the chain rule. All that is needed is that every operation knows its own derivative.
    TensorFlow implements this and anything using `tf.*` operations only can use this technique.

        Args:
            func: Function without arguments that depends on `params`
            params: Parameters that `func` implicitly depends on and with respect to which the
                derivatives will be taken.

        Returns:
            Hessian matrix
    """

    return automatic_value_gradients_hessian(func, params, hessian=hessian)[2]


def automatic_value_gradient_hessian(func: Callable = None, params: Iterable["zfit.Parameter"] = None,
                                     value_grad_func=None,
                                     hessian=None) -> [tf.Tensor, tf.Tensor, tf.Tensor]:
    """Calculate using autodiff the gradients and hessian matrix of `func()` wrt `params`; also return `func()`.

    Automatic differentiation (autodiff) is a way of retreiving the derivative of x wrt y. It works by consecutively
    applying the chain rule. All that is needed is that every operation knows its own derivative.
    TensorFlow implements this and anything using `tf.*` operations only can use this technique.

        Args:
            func: Function without arguments that depends on `params`
            params: Parameters that `func` implicitly depends on and with respect to which the
                derivatives will be taken.

        Returns:
            Value, gradient and hessian matrix
    """
    if params is None:
        raise ValueError("Parameters have to be specified, are currently None.")
    if func is None and value_grad_func is None:
        ValueError("Either `func` or `value_grad_func` has to be specified.")

    from .. import z
    persistant = hessian == 'diag' or tf.executing_eagerly()  # currently needed, TODO: can we better parallelize that?
    with tf.GradientTape(persistent=persistant, watch_accessed_variables=False) as tape:
        tape.watch(params)
        if callable(value_grad_func):
            loss, gradients = value_grad_func(params)
        else:
            loss, gradients = autodiff_value_gradients(func=func, params=params)
        if hessian != 'diag':
            gradients_tf = znp.stack(gradients)
    if hessian == 'diag':
        computed_hessian = znp.stack(
            [tape.gradient(grad, sources=param) for param, grad in zip(params, gradients)]
        )
    else:
        computed_hessian = z.convert_to_tensor(tape.jacobian(gradients_tf, sources=params,
                                                             experimental_use_pfor=False  # causes TF bug? Slow..
                                                             ))
    del tape
    return loss, gradients, computed_hessian


@deprecated(None, "Use `automatic_value_gradient_hessian` instead.")
def automatic_value_gradients_hessian(*args, **kwargs):
    return automatic_value_gradient_hessian(*args, **kwargs)

@tf.function(autograph=False)
def reduce_geometric_mean(input_tensor, axis=None, keepdims=False):
    log_mean = znp.mean(znp.log(input_tensor), axis=axis, keepdims=keepdims)
    return znp.exp(log_mean)


def log(x):
    x = _auto_upcast(x)
    return _auto_upcast(znp.log(x=x))



"""Baseclass for a Model.

Handle integration and sampling
"""

#  Copyright (c) 2021 zfit

import abc
import builtins
import contextlib
import inspect
import warnings
from collections import OrderedDict
from contextlib import suppress
from typing import Callable, Dict, List, Optional, Tuple, Union

import tensorflow as tf
from dotmap import DotMap
from tensorflow_probability.python import mcmc as mc

import zfit.z.numpy as znp

from .. import z
from ..core.integration import Integration
from ..settings import ztypes
from ..util import ztyping
from ..util.cache import GraphCachable
from ..util.exception import (AnalyticIntegralNotImplemented,
                              AnalyticSamplingNotImplemented,
                              BasePDFSubclassingError,
                              CannotConvertToNumpyError,
                              FunctionNotImplemented,
                              MultipleLimitsNotImplemented,
                              NormRangeNotImplemented, ShapeIncompatibleError,
                              SpaceIncompatibleError,
                              SpecificFunctionNotImplemented, SubclassingError,
                              WorkInProgressError)
from . import integration as zintegrate
from . import sample as zsample
from .baseobject import BaseNumeric
from .data import Data, SampleData, Sampler
from .dependents import _extract_dependencies
from .dimension import BaseDimensional
from .interfaces import ZfitData, ZfitModel, ZfitParameter, ZfitSpace
from .sample import UniformSampleAndWeights
from .space import Space, convert_to_space, no_norm_range, supports

_BaseModel_USER_IMPL_METHODS_TO_CHECK = {}


def _BaseModel_register_check_support(has_support: bool):
    """Marks a method that the subclass either *has* to or *can't* use the `@supports` decorator.

    Args:
        has_support: If True, flags that it **requires** the `@supports` decorator. If False,
            flags that the `@supports` decorator is **not allowed**.
    """
    if not isinstance(has_support, bool):
        raise TypeError("Has to be boolean.")

    def register(func):
        """Register a method to be checked to (if True) *has* `support` or (if False) has *no* `support`.

        Args:
            func:

        Returns:
            Function:
        """
        name = func.__name__
        _BaseModel_USER_IMPL_METHODS_TO_CHECK[name] = has_support
        func.__wrapped__ = _BaseModel_register_check_support
        return func

    return register


class BaseModel(BaseNumeric, GraphCachable, BaseDimensional, ZfitModel):
    """Base class for any generic model.

    # TODO instructions on how to use
    """
    _DEFAULTS_integration = DotMap()
    _DEFAULTS_integration.mc_sampler = lambda *args, **kwargs: mc.sample_halton_sequence(*args, randomized=False,
                                                                                         **kwargs)
    # _DEFAULTS_integration.mc_sampler = lambda dim, num_results, dtype: tf.random_uniform(maxval=1.,
    #                                                                                      shape=(num_results, dim),
    #                                                                                      dtype=dtype)
    _DEFAULTS_integration.draws_per_dim = 40000
    _DEFAULTS_integration.auto_numeric_integrator = zintegrate.auto_integrate

    _analytic_integral = None
    _inverse_analytic_integral = None
    _additional_repr = None

    def __init__(self, obs: ztyping.ObsTypeInput, params: Union[Dict[str, ZfitParameter], None] = None,
                 name: str = "BaseModel", dtype=ztypes.float,
                 **kwargs):
        """The base model to inherit from and overwrite `_unnormalized_pdf`.

        Args:
            dtype: the dtype of the model
            name: the name of the model
            params: A dictionary with the internal name of the parameter and
                the parameters itself the model depends on
        """
        super().__init__(name=name, dtype=dtype, params=params, **kwargs)
        self._check_set_space(obs)

        self._integration = DotMap()
        self._integration.auto_numeric_integrator = self._DEFAULTS_integration.auto_numeric_integrator
        self.integration = Integration(mc_sampler=self._DEFAULTS_integration.mc_sampler,
                                       draws_per_dim=self._DEFAULTS_integration.draws_per_dim)

        self._sample_and_weights = UniformSampleAndWeights

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        # check if subclass has decorator if required
        cls._subclass_check_support(methods_to_check=_BaseModel_USER_IMPL_METHODS_TO_CHECK,
                                    wrapper_not_overwritten=_BaseModel_register_check_support)
        cls._analytic_integral = zintegrate.AnalyticIntegral()
        cls._inverse_analytic_integral = []
        cls._additional_repr = {}

    @classmethod
    def _subclass_check_support(cls, methods_to_check, wrapper_not_overwritten):
        for method_name, has_support in methods_to_check.items():
            method = getattr(cls, method_name)
            if hasattr(method, "__wrapped__"):
                if method.__wrapped__ == wrapper_not_overwritten:
                    continue  # not overwritten, fine

            # here means: overwritten
            if hasattr(method, "__wrapped__"):
                if method.__wrapped__ == supports:
                    if has_support:
                        continue  # needs support, has been wrapped
                    else:
                        raise BasePDFSubclassingError("Method {} has been wrapped with supports "
                                                      "but is not allowed to. Has to handle all "
                                                      "arguments.".format(method_name))
                elif has_support:
                    raise BasePDFSubclassingError("Method {} has been overwritten and *has to* be "
                                                  "wrapped by `supports` decorator (don't forget () )"
                                                  "to call the decorator as it takes arguments"
                                                  "".format(method_name))
                elif not has_support:
                    continue  # no support, has not been wrapped with
            else:
                if not has_support:
                    continue  # not wrapped, no support, need no

            # if we reach this points, somethings was implemented wrongly
            raise BasePDFSubclassingError("Method {} has not been correctly wrapped with @supports "
                                          "OR has been wrapped but it should not be".format(method_name))

    # since subclasses can be funcs of pdfs, we need to now what to sample/integrate from
    @abc.abstractmethod
    def _func_to_integrate(self, x: ztyping.XType) -> tf.Tensor:
        raise NotImplementedError

    @abc.abstractmethod
    def _func_to_sample_from(self, x: ztyping.XType) -> Data:
        raise NotImplementedError

    @property
    def space(self) -> ZfitSpace:
        return self._space

    def _check_set_space(self, obs):
        if not isinstance(obs, ZfitSpace):
            obs = Space(obs=obs)
        self._check_n_obs(space=obs)
        self._space = obs.with_autofill_axes(overwrite=True)

    @contextlib.contextmanager
    def _convert_sort_x(self,
                        x: ztyping.XTypeInput,
                        partial: bool = False,
                        allow_none: bool = False) -> Data:
        if x is None:
            if not allow_none:
                raise ValueError(f"x {x} given to {self} must be non-empty (not None).")
            else:
                yield None

        elif isinstance(x, ZfitData):
            if x.obs is not None:
                with x.sort_by_obs(obs=self.obs, allow_superset=True):
                    yield x
            elif x.axes is not None:
                with x.sort_by_axes(axes=self.axes):
                    yield x
            else:
                assert False, "Neither the `obs` nor the `axes` are specified in `Data`"
        else:
            if not isinstance(x, (tf.Tensor, tf.Variable)):
                try:
                    x = z.convert_to_tensor(value=x)
                except TypeError:
                    raise TypeError(f"Wrong type of x ({type(x)}). Has to be a `Data` or convertible to a tf.Tensor")
            # check dimension
            x = self._add_dim_to_x(x=x)
            x_shape = x.shape.as_list()[-1]
            if not partial and x_shape != self.n_obs:
                raise ShapeIncompatibleError("The shape of x (={}) (in the last dim) does not"
                                             "match the shape (={})of the model".format(x_shape, self.n_obs))
            x = Data.from_tensor(obs=self.obs, tensor=x)
            yield x

    def _add_dim_to_x(self, x):  # TODO(Mayou36): remove function? unnecessary? dealt with in `Data`?
        if self.n_obs == 1:
            if len(x.shape.as_list()) == 0:
                x = znp.expand_dims(x, -1)
            if len(x.shape.as_list()) == 1:
                x = znp.expand_dims(x, -1)
        return x

    def update_integration_options(self, draws_per_dim=None, mc_sampler=None):
        """Set the integration options.

        Args:
            draws_per_dim: The draws for MC integration to do
            mc_sampler:
        """
        # mc_options = {} if mc_options is None else mc_options
        # numeric_options = {} if numeric_options is None else numeric_options
        # general_options = {} if general_options is None else general_options
        # analytic_options = {} if analytic_options is None else analytic_options
        # if analytic_options:
        #     raise NotImplementedError("analytic_options cannot be updated currently.")
        if draws_per_dim is not None:
            self.integration.draws_per_dim = draws_per_dim
        if mc_sampler is not None:
            self.integration.mc_sampler = mc_sampler

    # TODO: remove below? or add "analytic gradients"?
    def gradient(self, x: ztyping.XType, norm_range: ztyping.LimitsType, params: ztyping.ParamsTypeOpt = None):
        raise NotImplementedError("Are the gradients needed?")

    def _check_input_norm_range(self, norm_range, none_is_error=False) -> Optional[ZfitSpace]:
        """Convert to :py:class:`~zfit.Space`.

        Args:
            norm_range:
            none_is_error: if both `norm_range` and `self.norm_range` are None, the default
                value is `False` (meaning: no range specified-> no normalization to be done). If
                this is set to true, two `None` will raise a Value error.

        Returns:
            Union[:py:class:`~zfit.Space`, False]:
        """
        if norm_range is None or (isinstance(norm_range, ZfitSpace) and not norm_range.limits_are_set):
            if none_is_error:
                raise ValueError("Normalization range `norm_range` has to be specified or"
                                 "a default normalization range has to be set. Currently, both are None")

        return self.convert_sort_space(limits=norm_range)

    def _check_input_limits(self, limits, none_is_error=False):
        if limits is None or (isinstance(limits, ZfitSpace) and not limits.has_limits):
            if none_is_error:
                raise ValueError("The `limits` have to be specified and not be None")
            # else:
            #     limits = False

        return self.convert_sort_space(limits=limits)

    def convert_sort_space(self, obs: Union[ztyping.ObsTypeInput, ztyping.LimitsTypeInput] = None,
                           axes: ztyping.AxesTypeInput = None,
                           limits: ztyping.LimitsTypeInput = None) -> Union[ZfitSpace, None]:
        """Convert the inputs (using eventually `obs`, `axes`) to
        :py:class:`~zfit.ZfitSpace` and sort them according to own `obs`.

        Args:
            obs:
            axes:
            limits:

        Returns:
        """
        if obs is None:  # for simple limits to convert them
            obs = self.obs
        elif not set(obs).intersection(self.obs):
            raise SpaceIncompatibleError("The given space {obs} is not compatible with the obs of the pdfs{self.obs};"
                                         " they are disjoint.")
        space = convert_to_space(obs=obs, axes=axes, limits=limits)

        if self.space is not None:  # e.g. not the first call
            space = space.with_coords(self.space, allow_superset=True, allow_subset=True)
        return space

    # Integrals

    @_BaseModel_register_check_support(True)
    def _integrate(self, limits, norm_range):
        raise SpecificFunctionNotImplemented

    @z.function(wraps='model')
    def integrate(self,
                  limits: ztyping.LimitsType,
                  norm_range: ztyping.LimitsType = None,
                  *,
                  x: Optional[ztyping.DataInputType] = None) -> ztyping.XType:
        """Integrate the function over `limits` (normalized over `norm_range` if not False).

        Args:
            limits: the limits to integrate over
            norm_range: the limits to normalize over or False to integrate the
                unnormalized probability

        Returns:
            The integral value as a scalar with shape ()
        """
        norm_range = self._check_input_norm_range(norm_range)
        limits = self._check_input_limits(limits=limits)
        with self._convert_sort_x(x, allow_none=True) as x:
            integral = self._single_hook_integrate(limits=limits, norm_range=norm_range, x=x)
        # TODO: allow integral values as arrays?
        # if isinstance(integral, tf.Tensor):
        #     if not integral.shape.as_list() == []:
        #         raise ShapeIncompatibleError("Error in integral creation, should return an integral "
        #                                      "with shape () (resp. [] as list), current shape "
        #                                      "{}. If you registered an analytic integral which is used"
        #                                      "now, make sure to return a scalar and not a tensor "
        #                                      "(typically: shape is (1,) insead of () -> return tensor[0] "
        #                                      "instead of tensor)".format(integral.shape.as_list()))
        return integral

    def _single_hook_integrate(self, limits, norm_range, x):
        del x  # TODO HACK: how and what to pass through?
        return self._hook_integrate(limits=limits, norm_range=norm_range)

    def _hook_integrate(self, limits, norm_range):
        return self._norm_integrate(limits=limits, norm_range=norm_range)

    def _norm_integrate(self, limits, norm_range):
        try:
            integral = self._limits_integrate(limits=limits, norm_range=norm_range)
        except NormRangeNotImplemented:
            unnormalized_integral = self._limits_integrate(limits=limits, norm_range=False)
            normalization = self._limits_integrate(limits=norm_range, norm_range=False)
            integral = unnormalized_integral / normalization
        return integral

    def _limits_integrate(self, limits, norm_range):
        try:
            integral = self._call_integrate(limits=limits, norm_range=norm_range)
        except MultipleLimitsNotImplemented:
            integrals = []
            for sub_limits in limits:
                integrals.append(self._call_integrate(limits=sub_limits, norm_range=norm_range))
            integral = z.reduce_sum(znp.stack(integrals), axis=0)  # TODO: remove stack?
        return integral

    def _call_integrate(self, limits, norm_range):

        with suppress(FunctionNotImplemented):
            return self._integrate(limits=limits, norm_range=norm_range)
        with suppress(AnalyticIntegralNotImplemented):
            return self._hook_analytic_integrate(limits=limits, norm_range=norm_range)
        return self._fallback_integrate(limits=limits, norm_range=norm_range)

    def _fallback_integrate(self, limits, norm_range):
        axes = limits.axes
        max_axes = self._analytic_integral.get_max_axes(limits=limits, axes=axes)

        integral = None
        if max_axes and integral is None:  # TODO improve handling of available analytic integrals
            with suppress(AnalyticIntegralNotImplemented):
                def part_int(x):
                    """Temporary partial integration function."""
                    return self._hook_partial_analytic_integrate(x, limits=limits, norm_range=norm_range)

                integral = self._auto_numeric_integrate(func=part_int, limits=limits)
        if integral is None:
            integral = self._hook_numeric_integrate(limits=limits, norm_range=norm_range)
        return integral

    @classmethod
    def register_analytic_integral(cls, func: Callable, limits: ztyping.LimitsType = None,
                                   priority: Union[int, float] = 50, *,
                                   supports_norm_range: bool = False,
                                   supports_multiple_limits: bool = False) -> None:
        """Register an analytic integral with the class.

        Args:
            func: A function that calculates the (partial) integral over the axes `limits`.
                The signature has to be the following:

                    * x (:py:class:`~zfit.core.interfaces.ZfitData`, None): the data for the remaining axes in a partial
                        integral. If it is not a partial integral, this will be None.
                    * limits (:py:class:`~zfit.ZfitSpace`): the limits to integrate over.
                    * norm_range (:py:class:`~zfit.ZfitSpace`, None): Normalization range of the integral.
                        If not `supports_supports_norm_range`, this will be None.
                    * params (Dict[param_name, :py:class:`zfit.Parameters`]): The parameters of the model.
                    * model (:py:class:`~zfit.core.interfaces.ZfitModel`):The model that is being integrated.

            limits: |limits_arg_descr|
            priority: Priority of the function. If multiple functions cover the same space, the one with the
                highest priority will be used.
            supports_multiple_limits: If `True`, the `limits` given to the integration function can have
                multiple limits. If `False`, only simple limits will pass through and multiple limits will be
                auto-handled.
            supports_norm_range: If `True`, `norm_range` argument to the function may not be `None`.
                If `False`, `norm_range` will always be `None` and care is taken of the normalization automatically.
        """
        cls._analytic_integral.register(func=func, limits=limits, supports_norm_range=supports_norm_range,
                                        priority=priority, supports_multiple_limits=supports_multiple_limits)

    @classmethod
    def register_inverse_analytic_integral(cls, func: Callable) -> None:
        """Register an inverse analytical integral, the inverse (unnormalized) cdf.

        Args:
            func: A function with the signature `func(x, params)`, where `x` is a Data object
                and `params` is a dict.
        """
        if cls._inverse_analytic_integral:
            cls._inverse_analytic_integral[0] = func
        else:
            cls._inverse_analytic_integral.append(func)

    @_BaseModel_register_check_support(True)
    def _analytic_integrate(self, limits, norm_range):
        raise SpecificFunctionNotImplemented

    def analytic_integrate(self, limits: ztyping.LimitsType, norm_range: ztyping.LimitsType = None) -> ztyping.XType:
        """Analytical integration over function and raise Error if not possible.

        Args:
            limits: the limits to integrate over
            norm_range: the limits to normalize over

        Returns:
            The integral value
        Raises:
            AnalyticIntegralNotImplementedError: If no analytical integral is available (for this limits).
            NormRangeNotImplementedError: if the *norm_range* argument is not supported. This
                means that no analytical normalization is available, explicitly: the **analytical**
                integral over the limits = norm_range is not available.
        """
        norm_range = self._check_input_norm_range(norm_range)
        limits = self._check_input_limits(limits=limits)
        return self._single_hook_analytic_integrate(limits=limits, norm_range=norm_range)

    def _single_hook_analytic_integrate(self, limits, norm_range):
        return self._hook_analytic_integrate(limits=limits, norm_range=norm_range)

    def _hook_analytic_integrate(self, limits, norm_range):
        return self._norm_analytic_integrate(limits=limits, norm_range=norm_range)

    def _norm_analytic_integrate(self, limits, norm_range):
        try:
            integral = self._limits_analytic_integrate(limits=limits, norm_range=norm_range)
        except NormRangeNotImplemented:

            unnormalized_integral = self._limits_analytic_integrate(limits, norm_range=False)
            try:
                normalization = self._limits_analytic_integrate(limits=norm_range, norm_range=False)
            except (AnalyticIntegralNotImplemented):
                raise NormRangeNotImplemented("Function does not support this (or even any)"
                                              "normalization range 'norm_range'."
                                              " This usually means,that no analytic integral "
                                              "is available for this function. Due to rule "
                                              "safety, an analytical normalization has to "
                                              "be available and no attempt of numerical "
                                              "normalization was made.")
            else:
                integral = unnormalized_integral / normalization
        return integral

    def _limits_analytic_integrate(self, limits, norm_range):
        try:
            integral = self._call_analytic_integrate(limits, norm_range=norm_range)
        except MultipleLimitsNotImplemented:
            integrals = []
            for sub_limits in limits:
                integrals.append(self._call_analytic_integrate(limits=sub_limits, norm_range=norm_range))
            integral = z.reduce_sum(znp.stack(integrals), axis=0)
        return integral

    def _call_analytic_integrate(self, limits, norm_range):
        with suppress(FunctionNotImplemented, AnalyticIntegralNotImplemented):
            return self._analytic_integrate(limits=limits, norm_range=norm_range)
        return self._fallback_analytic_integrate(limits=limits, norm_range=norm_range)

    def _fallback_analytic_integrate(self, limits, norm_range):
        try:
            return self._analytic_integral.integrate(x=None, limits=limits, axes=limits.axes,
                                                     norm_range=norm_range, model=self, params=self.params)
        except (SpecificFunctionNotImplemented, AnalyticIntegralNotImplemented):
            raise AnalyticIntegralNotImplemented

    @_BaseModel_register_check_support(True)
    def _numeric_integrate(self, limits, norm_range):
        raise SpecificFunctionNotImplemented

    def numeric_integrate(self, limits: ztyping.LimitsType, norm_range: ztyping.LimitsType = None) -> ztyping.XType:
        """Numerical integration over the model.

        Args:
            limits: the limits to integrate over
            norm_range: the limits to normalize over

        Returns:
            The integral value
        """
        norm_range = self._check_input_norm_range(norm_range)
        limits = self._check_input_limits(limits=limits)

        return self._single_hook_numeric_integrate(limits=limits, norm_range=norm_range)

    def _single_hook_numeric_integrate(self, limits, norm_range):
        return self._hook_numeric_integrate(limits=limits, norm_range=norm_range)

    def _hook_numeric_integrate(self, limits, norm_range):
        return self._norm_numeric_integrate(limits=limits, norm_range=norm_range)

    def _norm_numeric_integrate(self, limits, norm_range):
        try:
            integral = self._limits_numeric_integrate(limits=limits, norm_range=norm_range)
        except NormRangeNotImplemented:
            assert not norm_range.limits_are_false, "Internal: the caught Error should not be raised."
            unnormalized_integral = self._limits_numeric_integrate(limits=limits, norm_range=False)
            normalization = self._limits_numeric_integrate(limits=norm_range, norm_range=False)
            integral = unnormalized_integral / normalization
        return integral

    def _limits_numeric_integrate(self, limits, norm_range):
        try:
            integral = self._call_numeric_integrate(limits=limits, norm_range=norm_range)
        except MultipleLimitsNotImplemented:
            integrals = []
            for sub_limits in limits:
                integrals.append(self._call_numeric_integrate(limits=sub_limits, norm_range=norm_range))
            integral = z.reduce_sum(znp.stack(integrals), axis=0)

        return integral

    def _call_numeric_integrate(self, limits, norm_range):
        with suppress(FunctionNotImplemented):
            return self._numeric_integrate(limits=limits, norm_range=norm_range)
        return self._fallback_numeric_integrate(limits=limits, norm_range=norm_range)

    def _fallback_numeric_integrate(self, limits, norm_range):
        return self._auto_numeric_integrate(func=self._func_to_integrate, limits=limits, norm_range=norm_range)

    @_BaseModel_register_check_support(True)
    def _partial_integrate(self, x, limits, norm_range):
        raise SpecificFunctionNotImplemented

    @z.function(wraps='model')
    def partial_integrate(self, x: ztyping.XTypeInput, limits: ztyping.LimitsType,
                          norm_range: ztyping.LimitsType = None) -> ztyping.XTypeReturn:
        """Partially integrate the function over the `limits` and evaluate it at `x`.

        Dimension of `limits` and `x` have to add up to the full dimension and be therefore equal
        to the dimensions of `norm_range` (if not False)

        Args:
            x: The value at which the partially integrated function will be evaluated
            limits: the limits to integrate over. Can contain only some axes
            norm_range: the limits to normalize over. Has to have all axes

        Returns:
            The value of the partially integrated function evaluated at `x`.
        """
        norm_range = self._check_input_norm_range(norm_range=norm_range)
        limits = self._check_input_limits(limits=limits)
        with self._convert_sort_x(x, partial=True) as x:
            return self._single_hook_partial_integrate(x=x, limits=limits, norm_range=norm_range)

    def _single_hook_partial_integrate(self, x, limits, norm_range):
        return self._hook_partial_integrate(x=x, limits=limits, norm_range=norm_range)

    def _hook_partial_integrate(self, x, limits, norm_range):
        integral = self._norm_partial_integrate(x=x, limits=limits, norm_range=norm_range)
        return integral

    def _norm_partial_integrate(self, x, limits, norm_range):
        try:
            integral = self._limits_partial_integrate(x=x, limits=limits, norm_range=norm_range)
        except NormRangeNotImplemented:
            assert not norm_range.limits_are_false, "Internal: the caught Error should not be raised."
            unnormalized_integral = self._limits_partial_integrate(x=x, limits=limits, norm_range=False)
            normalization = self._hook_integrate(limits=norm_range, norm_range=False)
            integral = unnormalized_integral / normalization
        return integral

    def _limits_partial_integrate(self, x, limits, norm_range):
        try:
            integral = self._call_partial_integrate(x=x, limits=limits, norm_range=norm_range)
        except MultipleLimitsNotImplemented:
            integrals = []
            for sub_limit in limits:
                integrals.append(self._call_partial_integrate(x=x, limits=sub_limit, norm_range=norm_range))
            integral = z.reduce_sum(znp.stack(integrals), axis=0)

        return integral

    def _call_partial_integrate(self, x, limits, norm_range):

        with suppress(FunctionNotImplemented):
            return self._partial_integrate(x=x, limits=limits, norm_range=norm_range)
        with suppress(AnalyticIntegralNotImplemented):
            return self._hook_partial_analytic_integrate(x=x, limits=limits, norm_range=norm_range)
        try:
            return self._fallback_partial_integrate(x=x, limits=limits, norm_range=norm_range)
        except FunctionNotImplemented:
            raise AnalyticIntegralNotImplemented

    def _fallback_partial_integrate(self, x, limits: ZfitSpace, norm_range: ZfitSpace):
        max_axes = self._analytic_integral.get_max_axes(limits=limits, axes=limits.axes)
        if max_axes:
            sublimits = limits.get_subspace(axes=max_axes)

            def part_int(x):  # change to partial integrate max axes?
                """Temporary partial integration function."""
                return self._hook_partial_analytic_integrate(x=x, limits=sublimits, norm_range=norm_range)

            axes = list(set(limits.axes) - set(max_axes))
            limits = limits.get_subspace(axes=axes)
        else:
            part_int = self._func_to_integrate

        assert limits.axes, "Internal Error! Axes should not be empty, maybe cleanup."
        integral_vals = self._auto_numeric_integrate(func=part_int, limits=limits, x=x, norm_range=norm_range)

        return integral_vals

    @_BaseModel_register_check_support(True)
    def _partial_analytic_integrate(self, x, limits, norm_range):
        raise SpecificFunctionNotImplemented

    @z.function(wraps='model')
    def partial_analytic_integrate(self, x: ztyping.XTypeInput, limits: ztyping.LimitsType,
                                   norm_range: ztyping.LimitsType = None) -> ztyping.XTypeReturn:
        """Do analytical partial integration of the function over the `limits` and evaluate it at `x`.

        Dimension of `limits` and `x` have to add up to the full dimension and be therefore equal
        to the dimensions of `norm_range` (if not False)

        Args:
            x: The value at which the partially integrated function will be evaluated
            limits: the limits to integrate over. Can contain only some axes
            norm_range: the limits to normalize over. Has to have all axes

        Returns:
            The value of the partially integrated function evaluated at `x`.

        Raises:
            AnalyticIntegralNotImplementedError: if the *analytic* integral (over this limits) is not implemented
            NormRangeNotImplementedError: if the *norm_range* argument is not supported. This
                means that no analytical normalization is available, explicitly: the **analytical**
                integral over the limits = norm_range is not available.
        """
        norm_range = self._check_input_norm_range(norm_range=norm_range)
        limits = self._check_input_limits(limits=limits)
        with self._convert_sort_x(x, partial=True) as x:
            return self._single_hook_partial_analytic_integrate(x=x, limits=limits, norm_range=norm_range)

    def _single_hook_partial_analytic_integrate(self, x, limits, norm_range):
        return self._hook_partial_analytic_integrate(x=x, limits=limits, norm_range=norm_range)

    def _hook_partial_analytic_integrate(self, x, limits, norm_range):
        return self._norm_partial_analytic_integrate(x=x, limits=limits, norm_range=norm_range)

    def _norm_partial_analytic_integrate(self, x, limits, norm_range):
        try:
            integral = self._limits_partial_analytic_integrate(x=x, limits=limits, norm_range=norm_range)
        except NormRangeNotImplemented:
            assert not norm_range.limits_are_false, "Internal: the caught Error should not be raised."
            unnormalized_integral = self._limits_partial_analytic_integrate(x=x, limits=limits, norm_range=False)
            try:
                normalization = self._limits_analytic_integrate(limits=norm_range, norm_range=False)
            except AnalyticIntegralNotImplemented:
                raise NormRangeNotImplemented("Function does not support this (or even any) normalization range"
                                              " 'norm_range'. This usually means,that no analytic integral "
                                              "is available for this function. An analytical normalization has to "
                                              "be available and no attempt of numerical normalization was made.")
            else:
                integral = unnormalized_integral / normalization
        return integral

    def _limits_partial_analytic_integrate(self, x, limits, norm_range):
        try:
            integral = self._call_partial_analytic_integrate(x=x, limits=limits, norm_range=norm_range)
        except MultipleLimitsNotImplemented:
            integrals = []
            for sub_limits in limits:
                integrals.append(self._call_partial_analytic_integrate(x=x, limits=sub_limits, norm_range=norm_range))
            integral = z.reduce_sum(znp.stack(integrals), axis=0)

        return integral

    def _call_partial_analytic_integrate(self, x, limits, norm_range):
        with suppress(FunctionNotImplemented, AnalyticIntegralNotImplemented):
            return self._partial_analytic_integrate(x=x, limits=limits, norm_range=norm_range)
        return self._fallback_partial_analytic_integrate(x=x, limits=limits, norm_range=norm_range)

    def _fallback_partial_analytic_integrate(self, x, limits, norm_range):
        try:
            return self._analytic_integral.integrate(x=x, limits=limits, axes=limits.axes,
                                                     norm_range=norm_range, model=self, params=self.params)
        except (SpecificFunctionNotImplemented, AnalyticIntegralNotImplemented):
            raise AnalyticIntegralNotImplemented

    @_BaseModel_register_check_support(True)
    def _partial_numeric_integrate(self, x, limits, norm_range):
        raise SpecificFunctionNotImplemented

    @z.function(wraps='model')
    def partial_numeric_integrate(self, x: ztyping.XType, limits: ztyping.LimitsType,
                                  norm_range: ztyping.LimitsType = None) -> ztyping.XType:
        """Force numerical partial integration of the function over the `limits` and evaluate it at `x`.

        Dimension of `limits` and `x` have to add up to the full dimension and be therefore equal
        to the dimensions of `norm_range` (if not False)

        Args:
            x: The value at which the partially integrated function will be evaluated
            limits: the limits to integrate over. Can contain only some axes
            norm_range: the limits to normalize over. Has to have all axes

        Returns:
            The value of the partially integrated function evaluated at `x`.
        """
        norm_range = self._check_input_norm_range(norm_range)
        limits = self._check_input_limits(limits=limits)
        with self._convert_sort_x(x, partial=True) as x:
            return self._single_hook_partial_numeric_integrate(x=x, limits=limits, norm_range=norm_range)

    def _single_hook_partial_numeric_integrate(self, x, limits, norm_range):
        return self._hook_partial_numeric_integrate(x=x, limits=limits, norm_range=norm_range)

    def _hook_partial_numeric_integrate(self, x, limits, norm_range):
        integral = self._norm_partial_numeric_integrate(x=x, limits=limits, norm_range=norm_range)
        return integral

    def _norm_partial_numeric_integrate(self, x, limits, norm_range):
        try:
            integral = self._limits_partial_numeric_integrate(x=x, limits=limits, norm_range=norm_range)
        except NormRangeNotImplemented:
            assert not norm_range.limits_are_false, "Internal: the caught Error should not be raised."
            unnormalized_integral = self._limits_partial_numeric_integrate(x=x, limits=limits, norm_range=False)
            integral = unnormalized_integral / self._hook_numeric_integrate(limits=norm_range, norm_range=norm_range)
        return integral

    def _limits_partial_numeric_integrate(self, x, limits, norm_range):
        try:
            integral = self._call_partial_numeric_integrate(x=x, limits=limits, norm_range=norm_range)
        except MultipleLimitsNotImplemented:
            integrals = []
            for sub_limits in limits:
                integrals.append(self._call_partial_numeric_integrate(x=x, limits=sub_limits, norm_range=norm_range))
            integral = z.reduce_sum(znp.stack(integrals), axis=0)
        return integral

    def _call_partial_numeric_integrate(self, x, limits, norm_range):
        with suppress(SpecificFunctionNotImplemented):
            return self._partial_numeric_integrate(x=x, limits=limits, norm_range=norm_range)
        return self._fallback_partial_numeric_integrate(x=x, limits=limits, norm_range=norm_range)

    def _fallback_partial_numeric_integrate(self, x, limits, norm_range=False):
        return self._auto_numeric_integrate(func=self._func_to_integrate, limits=limits, norm_range=norm_range, x=x)

    @no_norm_range
    @z.function(wraps='model')
    def _auto_numeric_integrate(self, func, limits, x=None, norm_range=False, **overwrite_options):
        integration_options = dict(func=func, limits=limits, n_axes=limits.n_obs, x=x, norm_range=norm_range,
                                   # auto from self
                                   dtype=self.dtype,
                                   mc_sampler=self.integration.mc_sampler,
                                   mc_options={
                                       "draws_per_dim": self.integration.draws_per_dim},
                                   **overwrite_options)
        return self._integration.auto_numeric_integrator(**integration_options)

    @supports()
    def _inverse_analytic_integrate(self, x):
        if not self._inverse_analytic_integral:
            raise AnalyticSamplingNotImplemented
        else:
            icdf = self._inverse_analytic_integral[0]
            params = inspect.signature(icdf).parameters
            if len(params) == 2:
                return icdf(x=x, params=self.params)
            elif len(params) == 3:
                return icdf(x=x, params=self.params, model=self)
            else:
                raise RuntimeError(f"icdf function does not have the right signature: {icdf}")

    def create_sampler(self, n: ztyping.nSamplingTypeIn = None, limits: ztyping.LimitsType = None,
                       fixed_params: Union[bool, List[ZfitParameter], Tuple[ZfitParameter]] = True) -> "Sampler":
        """Create a :py:class:`Sampler` that acts as `Data` but can be resampled, also with changed parameters and n.

            If `limits` is not specified, `space` is used (if the space contains limits).
            If `n` is None and the model is an extended pdf, 'extended' is used by default.


        Args:
            n: The number of samples to be generated. Can be a Tensor that will be
                or a valid string. Currently implemented:

                    - 'extended': samples `poisson(yield)` from each pdf that is extended.

            limits: From which space to sample.
            fixed_params: A list of `Parameters` that will be fixed during several `resample` calls.
                If True, all are fixed, if False, all are floating. If a :py:class:`~zfit.Parameter` is not fixed and
                its
                value gets updated (e.g. by a `Parameter.set_value()` call), this will be reflected in
                `resample`. If fixed, the Parameter will still have the same value as the `Sampler` has
                been created with when it resamples.

        Returns:
            :py:class:`~zfit.core.data.Sampler`

        Raises:
            NotExtendedPDFError: if 'extended' is chosen (implicitly by default or explicitly) as an
                option for `n` but the pdf itself is not extended.
            ValueError: if n is an invalid string option.
            InvalidArgumentError: if n is not specified and pdf is not extended.
        """

        limits = self._check_input_limits(limits=limits)

        if not limits.limits_are_set:
            limits = self.space  # TODO(Mayou36): clean up, better norm_range?
            if not limits.has_limits:
                raise ValueError("limits are False/None, have to be specified")

        if fixed_params is True:
            fixed_params = list(self.get_cache_deps(only_floating=False))
        elif fixed_params is False:
            fixed_params = []
        elif not isinstance(fixed_params, (list, tuple)):
            raise TypeError("`Fixed_params` has to be a list, tuple or a boolean.")

        def sample_func(n=n):
            return self._create_sampler_tensor(limits=limits, n=n)

        sample_data = Sampler.from_sample(sample_func=sample_func, n=n, obs=limits, fixed_params=fixed_params,
                                          dtype=self.dtype)

        return sample_data

    @z.function(wraps='model')
    def _create_sampler_tensor(self, limits, n):

        sample = self._single_hook_sample(n=n, limits=limits, x=None)
        return sample

    @_BaseModel_register_check_support(True)
    def _sample(self, n, limits: ZfitSpace):
        raise SpecificFunctionNotImplemented

    def sample(self, n: ztyping.nSamplingTypeIn = None,
               limits: ztyping.LimitsType = None,
               x: Optional[
                   ztyping.DataInputType] = None) -> SampleData:  # TODO: change poissonian top-level with multinomial
        """Sample `n` points within `limits` from the model.

        If `limits` is not specified, `space` is used (if the space contains limits).
        If `n` is None and the model is an extended pdf, 'extended' is used by default.

        Args:
            n: The number of samples to be generated. Can be a Tensor that will be
                or a valid string. Currently implemented:

                    - 'extended': samples `poisson(yield)` from each pdf that is extended.
            limits: In which region to sample in

        Returns:
            SampleData(n_obs, n_samples)

        Raises:
            NotExtendedPDFError: if 'extended' is (implicitly by default or explicitly) chosen as an
                option for `n` but the pdf itself is not extended.
            ValueError: if n is an invalid string option.
            InvalidArgumentError: if n is not specified and pdf is not extended.
        """
        if not isinstance(n, str):
            n = tf.convert_to_tensor(n)
            n = tf.cast(n, dtype=tf.int32)

        limits = self._check_input_limits(limits=limits)
        if not limits.limits_are_set:
            limits = self.space
            if not limits.has_limits:
                raise tf.errors.InvalidArgumentError("limits are False/None, have to be specified")
        limits = self._check_input_limits(limits=limits, none_is_error=True)

        @z.function(wraps='model_sampling')
        def run_tf(n, limits, x):
            sample = self._single_hook_sample(n=n, limits=limits, x=x)
            return sample

        with self._convert_sort_x(x, allow_none=True) as x:
            if x is not None:
                new_obs = limits * x.data_range
            else:
                new_obs = limits
            sample_data = SampleData.from_sample(sample=run_tf(n=n, limits=limits, x=x),
                                                 obs=new_obs)  # TODO: which limits?

        return sample_data

    def _single_hook_sample(self, n, limits, x=None):
        del x
        return self._hook_sample(n=n, limits=limits)

    def _hook_sample(self, limits, n):
        return self._norm_sample(n=n, limits=limits)

    def _norm_sample(self, n, limits):
        """Dummy function."""
        return self._limits_sample(n=n, limits=limits)

    def _limits_sample(self, n, limits):
        try:
            return self._call_sample(n=n, limits=limits)
        except MultipleLimitsNotImplemented as error:
            try:
                total_integral = self.analytic_integrate(limits, norm_range=False)
                sub_integrals = znp.concatenate([self.analytic_integrate(limit, norm_range=False) for limit in limits],
                                          axis=0)
            except AnalyticIntegralNotImplemented:
                raise MultipleLimitsNotImplemented("Cannot autohandle multiple limits as the analytic"
                                                   " integral is not available.") from error
            fracs = sub_integrals / total_integral
            n_samples = tf.unstack(z.random.counts_multinomial(n, probs=fracs), axis=0)

            samples = []
            for limit, n_sample in zip(limits, n_samples):
                sub_sample = self._call_sample(n=n_sample, limits=limit)
                if isinstance(sub_sample, ZfitData):
                    sub_sample = sub_sample.value()
                samples.append(sub_sample)
            sample = znp.concatenate(samples, axis=0)

        return sample

    def _call_sample(self, n, limits):
        with suppress(SpecificFunctionNotImplemented):
            return self._sample(n=n, limits=limits)
        with suppress(SpecificFunctionNotImplemented, AnalyticSamplingNotImplemented):
            return self._analytic_sample(n=n, limits=limits)
        return self._fallback_sample(n=n, limits=limits)

    def _analytic_sample(self, n, limits: ZfitSpace):  # TODO(Mayou36) implement multiple limits sampling
        if not self._inverse_analytic_integral:
            raise AnalyticSamplingNotImplemented  # TODO(Mayou36): create proper analytic sampling
        if limits.n_limits > 1:
            raise AnalyticSamplingNotImplemented
        try:
            lower_bound, upper_bound = limits.rect_limits_np
        except CannotConvertToNumpyError as err:
            raise WorkInProgressError("Currently, analytic sampling with Tensors not supported."
                                      " Needs implementation of analytic integrals with Tensors.") from err
        neg_infinities = (tuple((-float("inf"),) * limits.n_obs),)  # py34 change float("inf") to math.inf
        # to the cdf to get the limits for the inverse analytic integral
        try:
            lower_prob_lim = self._norm_analytic_integrate(limits=Space(limits=(neg_infinities,
                                                                                (lower_bound,)),
                                                                        axes=limits.axes),
                                                           norm_range=False)

            upper_prob_lim = self._norm_analytic_integrate(limits=Space(limits=(neg_infinities,
                                                                                (upper_bound,)),
                                                                        axes=limits.axes),
                                                           norm_range=False)
        except (SpecificFunctionNotImplemented, AnalyticIntegralNotImplemented):
            raise AnalyticSamplingNotImplemented(f"analytic sampling not possible because the analytic integral"
                                                 f" is not"" implemented for the boundaries: {limits}")
        prob_sample = z.random.uniform(shape=(n, limits.n_obs), minval=lower_prob_lim,
                                       maxval=upper_prob_lim)
        # with self._convert_sort_x(prob_sample) as x:
        x = prob_sample
        sample = self._inverse_analytic_integrate(x=x)
        return sample

    def _fallback_sample(self, n, limits):
        sample = zsample.accept_reject_sample(prob=self._func_to_sample_from, n=n, limits=limits,
                                              prob_max=None, dtype=self.dtype,
                                              sample_and_weights_factory=self._sample_and_weights)
        return sample

    @classmethod
    def register_additional_repr(cls, **kwargs):
        """Register an additional attribute to add to the repr.

        Args:
            any keyword argument. The value has to be gettable from the instance (has to be an
            attribute or callable method of self.
        """
        if cls._additional_repr is None:
            cls._additional_repr = {}
        overwritten_keys = set(kwargs).intersection(cls._additional_repr)
        if overwritten_keys:
            warnings.warn("The following keys have been overwritten while registering additional repr:"
                          "\n{}".format([str(k) for k in overwritten_keys]))
        cls._additional_repr = dict(cls._additional_repr, **kwargs)

    def _get_additional_repr(self, sorted=True):

        # nice name change
        sorted_ = sorted
        sorted = builtins.sorted
        # nice name change end

        additional_repr = OrderedDict()
        for key, val in self._additional_repr.items():
            try:
                new_obj = getattr(self, val)
            except AttributeError:
                raise AttributeError("The attribute {} is not a valid attribute of this class {}."
                                     "Cannot use it in __repr__. It was added using the"
                                     "`register_additional_repr` function".format(val, type(self)))
            else:
                if callable(new_obj):
                    new_obj = new_obj()
            additional_repr[key] = new_obj
        if sorted_:
            additional_repr = OrderedDict(sorted(additional_repr))
        return additional_repr

    def __repr__(self):  # TODO(mayou36):repr to baseobject with _repr

        return ("<zfit.{type_name} "
                " params=[{params}]".format(type_name=type(self),
                                            params=", ".join(sorted(str(p.name) for p in self.params.values())),
                                            ))

    def _check_input_x_function(self, func):
        # TODO: signature etc?
        if not callable(func):
            raise TypeError("Function {} is not callable.")
        return func

    def _get_dependencies(self) -> ztyping.DependentsType:
        return _extract_dependencies(self.get_params())

    def __add__(self, other):
        from . import operations
        return operations.add(self, other)

    def __radd__(self, other):
        from . import operations
        return operations.add(other, self)

    def __mul__(self, other):
        from . import operations
        return operations.multiply(self, other)

    def __rmul__(self, other):
        from . import operations
        return operations.multiply(other, self)


class SimpleModelSubclassMixin:
    """Subclass a model: implement the corresponding function and specify _PARAMS.

    In order to create a custom model, two things have to be implemented: the class attribute
    _PARAMS has to be a list containing the names of the parameters and the corresponding
    function (_unnormalized_pdf/_func) has to be overridden.

    Example:

    .. code:: python

        class MyPDF(zfit.pdf.ZPDF):
            _PARAMS = ['mu', 'sigma']

            def _unnormalized_pdf(self, x):
                mu = self.params['mu']
                sigma = self.params['sigma']
                x = z.unstack_x(x)
                return z.exp(-z.square((x - mu) / sigma))
    """

    def __init__(self, *args, **kwargs):
        try:
            params = OrderedDict((name, kwargs.pop(name)) for name in self._PARAMS)
        except KeyError:
            raise ValueError("The following parameters are not given (as keyword arguments): {}"
                             "".format([k for k in self._PARAMS if k not in kwargs]))
        super().__init__(params=params, *args, **kwargs)
        # super().__init__(params=params, *args, **kwargs)  # use if upper fails

    @classmethod
    def _check_simple_model_subclass(cls):
        try:
            params = cls._PARAMS
        except AttributeError:
            raise SubclassingError("Need to define `_PARAMS` in the definition of the subclass."
                                   "Example:"
                                   "class MyModel(ZPDF):"
                                   "    _PARAMS = ['mu', 'sigma']")
        not_str = [param for param in params if not isinstance(param, str)]
        if not_str:
            raise TypeError(f"The following parameters are not strings in `_PARAMS`: ")



#  Copyright (c) 2021 zfit

class Singleton:
    __instance = None

    def __new__(cls, *args, **kwargs):
        instance = cls.__instance
        if instance is None:
            instance = super().__new__(cls)
            cls.__instance = instance

        return instance


class NotSpecified(Singleton):

    def __bool__(self):
        return False

    def __str__(self):
        return "A not specified state"


NONE = NotSpecified()


class ZfitNotImplemented:
    def __new__(cls, *args, **kwargs):
        raise RuntimeError("Cannot create an instance of it, meant to be used as a single object reference.")



#  Copyright (c) 2021 zfit

import numpy as np
import pytest

import zfit
from zfit import z
from zfit.util.exception import SubclassingError


def test_pdf_simple_subclass():
    class SimpleGauss(zfit.pdf.ZPDF):
        _PARAMS = ['mu', 'sigma']

        def _unnormalized_pdf(self, x):
            mu = self.params['mu']
            sigma = self.params['sigma']
            x = z.unstack_x(x)
            return z.exp(-z.square((x - mu) / sigma))

    gauss1 = SimpleGauss(obs='obs1', mu=3, sigma=5)

    prob = gauss1.pdf(np.random.random(size=(10, 1)), norm_range=(-4, 5))
    prob.numpy()

    with pytest.raises(ValueError):
        gauss2 = SimpleGauss('obs1', 1., sigma=5.)
    with pytest.raises(SubclassingError):
        class SimpleGauss2(zfit.pdf.ZPDF):

            def _unnormalized_pdf(self, x):
                mu = self.params['mu']
                sigma = self.params['sigma']
                x = z.unstack_x(x)
                return z.exp(-z.square((x - mu) / sigma))


def test_func_simple_subclass():
    class SimpleGaussFunc(zfit.func.ZFunc):
        _PARAMS = ['mu', 'sigma']

        def _func(self, x):
            mu = self.params['mu']
            sigma = self.params['sigma']
            x = z.unstack_x(x)
            return z.exp(-z.square((x - mu) / sigma))

    gauss1 = SimpleGaussFunc(obs='obs1', mu=3, sigma=5)

    value = gauss1.func(np.random.random(size=(10, 1)))

    with pytest.raises(ValueError):
        gauss2 = SimpleGaussFunc('obs1', 1., sigma=5.)
    with pytest.raises(SubclassingError):
        class SimpleGauss2(zfit.func.ZFunc):

            def _func(self, x):
                mu = self.params['mu']
                sigma = self.params['sigma']
                x = z.unstack_x(x)
                return z.exp(-z.square((x - mu) / sigma))



#  Copyright (c) 2021 zfit
import numpy as np
import pytest
import tensorflow as tf

import zfit.core.basepdf
import zfit.models.dist_tfp
import zfit.settings
from zfit import z
from zfit.core.loss import UnbinnedNLL
from zfit.core.space import Space
from zfit.minimize import Minuit
from zfit.pdf import Gauss
from zfit.util.exception import (BehaviorUnderDiscussion,
                                 IntentionAmbiguousError)

mu_true = 1.2
sigma_true = 4.1
mu_true2 = 1.01
sigma_true2 = 3.5

yield_true = 3000
test_values_np = np.random.normal(loc=mu_true, scale=sigma_true, size=(yield_true, 1))


def create_test_values(size):
    return tf.random.normal(mean=mu_true, stddev=sigma_true, shape=(size, 1))


test_values_np2 = np.random.normal(loc=mu_true2, scale=sigma_true2, size=yield_true)

low, high = -24.3, 28.6


def create_params1(nameadd=""):
    mu1 = zfit.Parameter("mu1" + nameadd, z.to_real(mu_true) - 0.2, mu_true - 1., mu_true + 1.)
    sigma1 = zfit.Parameter("sigma1" + nameadd, z.to_real(sigma_true) - 0.3, sigma_true - 2., sigma_true + 2.)
    return mu1, sigma1


def create_params2(nameadd=""):
    mu2 = zfit.Parameter("mu25" + nameadd, z.to_real(mu_true) - 0.2, mu_true - 1., mu_true + 1.)
    sigma2 = zfit.Parameter("sigma25" + nameadd, z.to_real(sigma_true) - 0.3, sigma_true - 2., sigma_true + 2.)
    return mu2, sigma2


def create_params3(nameadd=""):
    mu3 = zfit.Parameter("mu35" + nameadd, z.to_real(mu_true) - 0.2, mu_true - 1., mu_true + 1.)
    sigma3 = zfit.Parameter("sigma35" + nameadd, z.to_real(sigma_true) - 0.3, sigma_true - 2., sigma_true + 2.)
    yield3 = zfit.Parameter("yield35" + nameadd, yield_true + 300, 0, 10000000)
    return mu3, sigma3, yield3


obs1 = zfit.Space('obs1', (np.min([test_values_np[:, 0], test_values_np2]) - 1.4,
                           np.max([test_values_np[:, 0], test_values_np2]) + 2.4))

mu_constr = [1.6, 0.02]  # mu, sigma
sigma_constr = [3.5, 0.01]
constr = lambda: [mu_constr[1], sigma_constr[1]]
constr_tf = lambda: z.convert_to_tensor(constr())
covariance = lambda: np.array([[mu_constr[1] ** 2, 0], [0, sigma_constr[1] ** 2]])
covariance_tf = lambda: z.convert_to_tensor(covariance())


def create_gauss1():
    mu, sigma = create_params1()
    return Gauss(mu, sigma, obs=obs1, name="gaussian1"), mu, sigma


def create_gauss2():
    mu, sigma = create_params2()
    return Gauss(mu, sigma, obs=obs1, name="gaussian2"), mu, sigma


def create_gauss3ext():
    mu, sigma, yield3 = create_params3()
    gaussian3 = Gauss(mu, sigma, obs=obs1, name="gaussian3")
    gaussian3 = gaussian3.create_extended(yield3)
    return gaussian3, mu, sigma, yield3


def create_simpel_loss():
    _, mu1, sigma1 = create_gauss1()
    return zfit.loss.SimpleLoss(lambda x: (x[0] - 0.37) ** 4 * (x[1] - 2.34) ** 4,
                                params=[mu1, sigma1],
                                errordef=0.5)


def create_simultaneous_loss():
    test_values = tf.constant(test_values_np)
    test_values = zfit.Data.from_tensor(obs=obs1, tensor=test_values)
    test_values2 = tf.constant(test_values_np2)
    test_values2 = zfit.Data.from_tensor(obs=obs1, tensor=test_values2)
    gaussian1, mu1, sigma1 = create_gauss1()
    gaussian2, mu2, sigma2 = create_gauss2()
    gaussian2 = gaussian2.create_extended(zfit.Parameter('yield_gauss2', 5))
    nll = zfit.loss.UnbinnedNLL(model=[gaussian1, gaussian2],
                                data=[test_values, test_values2],
                                )
    return mu1, mu2, nll, sigma1, sigma2


@pytest.mark.parametrize('size', [None, 5000, 50000, 300000, 3_000_000])
@pytest.mark.flaky(3)  # minimization can fail
def test_extended_unbinned_nll(size):
    if size is None:
        test_values = z.constant(test_values_np)
        size = test_values.shape[0]
    else:
        test_values = create_test_values(size)
    test_values = zfit.Data.from_tensor(obs=obs1, tensor=test_values)
    gaussian3, mu3, sigma3, yield3 = create_gauss3ext()
    nll = zfit.loss.ExtendedUnbinnedNLL(model=gaussian3,
                                        data=test_values,
                                        fit_range=(-20, 20))
    assert {mu3, sigma3, yield3} == nll.get_params()
    minimizer = Minuit(tol=1e-4)
    status = minimizer.minimize(loss=nll)
    params = status.params
    assert params[mu3]['value'] == pytest.approx(zfit.run(tf.math.reduce_mean(test_values)), rel=0.05)
    assert params[sigma3]['value'] == pytest.approx(zfit.run(tf.math.reduce_std(test_values)), rel=0.05)
    assert params[yield3]['value'] == pytest.approx(size, rel=0.005)


def test_unbinned_simultaneous_nll():
    mu1, mu2, nll, sigma1, sigma2 = create_simultaneous_loss()
    minimizer = Minuit(tol=1e-5)
    status = minimizer.minimize(loss=nll, params=[mu1, sigma1, mu2, sigma2])
    params = status.params
    assert set(nll.get_params()) == {mu1, mu2, sigma1, sigma2}

    assert params[mu1]['value'] == pytest.approx(np.mean(test_values_np), rel=0.007)
    assert params[mu2]['value'] == pytest.approx(np.mean(test_values_np2), rel=0.007)
    assert params[sigma1]['value'] == pytest.approx(np.std(test_values_np), rel=0.007)
    assert params[sigma2]['value'] == pytest.approx(np.std(test_values_np2), rel=0.007)


@pytest.mark.flaky(3)
@pytest.mark.parametrize('weights', (None, np.random.normal(loc=1., scale=0.2, size=test_values_np.shape[0])))
@pytest.mark.parametrize('sigma', (constr, constr_tf, covariance, covariance_tf))
@pytest.mark.parametrize('options', ({'subtr_const': False}, {'subtr_const': True}))
def test_unbinned_nll(weights, sigma, options):
    gaussian1, mu1, sigma1 = create_gauss1()
    gaussian2, mu2, sigma2 = create_gauss2()

    test_values = tf.constant(test_values_np)
    test_values = zfit.Data.from_tensor(obs=obs1, tensor=test_values, weights=weights)
    nll_object = zfit.loss.UnbinnedNLL(model=gaussian1, data=test_values, options=options)
    minimizer = Minuit(tol=1e-5)
    status = minimizer.minimize(loss=nll_object, params=[mu1, sigma1])
    params = status.params
    rel_error = 0.12 if weights is None else 0.1  # more fluctuating with weights

    assert params[mu1]['value'] == pytest.approx(np.mean(test_values_np), rel=rel_error)
    assert params[sigma1]['value'] == pytest.approx(np.std(test_values_np), rel=rel_error)

    constraints = zfit.constraint.nll_gaussian(params=[mu2, sigma2],
                                               observation=[mu_constr[0], sigma_constr[0]],
                                               uncertainty=sigma())
    nll_object = UnbinnedNLL(model=gaussian2, data=test_values,
                             constraints=constraints, options=options)

    minimizer = Minuit(tol=1e-4)
    status = minimizer.minimize(loss=nll_object, params=[mu2, sigma2])
    params = status.params
    if weights is None:
        assert params[mu2]['value'] > np.average(test_values_np, weights=weights)
        assert params[sigma2]['value'] < np.std(test_values_np)


def test_add():
    param1 = zfit.Parameter("param1", 1.)
    param2 = zfit.Parameter("param2", 2.)
    param3 = zfit.Parameter("param3", 2.)

    pdfs = [0] * 4
    pdfs[0] = Gauss(param1, 4, obs=obs1)
    pdfs[1] = Gauss(param2, 5, obs=obs1)
    pdfs[2] = Gauss(3, 6, obs=obs1)
    pdfs[3] = Gauss(4, 7, obs=obs1)

    datas = [0] * 4
    datas[0] = zfit.Data.from_tensor(obs=obs1, tensor=z.constant(1.))
    datas[1] = zfit.Data.from_tensor(obs=obs1, tensor=z.constant(2.))
    datas[2] = zfit.Data.from_tensor(obs=obs1, tensor=z.constant(3.))
    datas[3] = zfit.Data.from_tensor(obs=obs1, tensor=z.constant(4.))

    ranges = [0] * 4
    ranges[0] = (1, 4)
    ranges[1] = Space(limits=(2, 5), obs=obs1)
    ranges[2] = Space(limits=(3, 6), obs=obs1)
    ranges[3] = Space(limits=(4, 7), obs=obs1)

    constraint1 = zfit.constraint.nll_gaussian(params=param1, observation=1., uncertainty=0.5)
    constraint2 = zfit.constraint.nll_gaussian(params=param3, observation=2., uncertainty=0.25)
    merged_contraints = [constraint1, constraint2]

    nll1 = UnbinnedNLL(model=pdfs[0], data=datas[0], fit_range=ranges[0], constraints=constraint1)
    nll2 = UnbinnedNLL(model=pdfs[1], data=datas[1], fit_range=ranges[1], constraints=constraint2)
    nll3 = UnbinnedNLL(model=[pdfs[2], pdfs[3]], data=[datas[2], datas[3]], fit_range=[ranges[2], ranges[3]])

    simult_nll = nll1 + nll2 + nll3

    assert simult_nll.model == pdfs
    assert simult_nll.data == datas

    ranges[0] = Space(limits=ranges[0], obs='obs1',
                      axes=(0,))  # for comparison, Space can only compare with Space
    ranges[1].coords._axes = (0,)
    ranges[2].coords._axes = (0,)
    ranges[3].coords._axes = (0,)
    assert simult_nll.fit_range == ranges

    def eval_constraint(constraints):
        return z.reduce_sum([c.value() for c in constraints]).numpy()

    assert eval_constraint(simult_nll.constraints) == eval_constraint(merged_contraints)
    assert set(simult_nll.get_params()) == {param1, param2, param3}


# @pytest.mark.xfail  # TODO(TF2): grads not supported, use numerical ones? Or calculate on the fly?
@pytest.mark.parametrize("chunksize", [10000000, 1000])
def test_gradients(chunksize):
    from numdifftools import Gradient
    zfit.run.chunking.active = True
    zfit.run.chunking.max_n_points = chunksize

    initial1 = 1.
    initial2 = 2
    param1 = zfit.Parameter("param1", initial1)
    param2 = zfit.Parameter("param2", initial2)

    gauss1 = Gauss(param1, 4, obs=obs1)
    gauss1.set_norm_range((-5, 5))
    gauss2 = Gauss(param2, 5, obs=obs1)
    gauss2.set_norm_range((-5, 5))

    data1 = zfit.Data.from_tensor(obs=obs1, tensor=z.constant(1., shape=(100,)))
    data1.set_data_range((-5, 5))
    data2 = zfit.Data.from_tensor(obs=obs1, tensor=z.constant(1., shape=(100,)))
    data2.set_data_range((-5, 5))

    nll = UnbinnedNLL(model=[gauss1, gauss2], data=[data1, data2])

    def loss_func(values):
        for val, param in zip(values, nll.get_cache_deps(only_floating=True)):
            param.set_value(val)
        return nll.value().numpy()

    # theoretical, numerical = tf.test.compute_gradient(loss_func, list(params))
    gradient1 = nll.gradient(params=param1)
    gradient_func = Gradient(loss_func)
    # gradient_func = lambda *args, **kwargs: list(gradient_func_numpy(*args, **kwargs))
    assert gradient1[0].numpy() == pytest.approx(gradient_func([param1.numpy()]))
    param1.set_value(initial1)
    param2.set_value(initial2)
    params = [param2, param1]
    gradient2 = nll.gradient(params=params)
    both_gradients_true = list(reversed(list(gradient_func([initial1, initial2]))))  # because param2, then param1
    assert [g.numpy() for g in gradient2] == pytest.approx(both_gradients_true)

    param1.set_value(initial1)
    param2.set_value(initial2)
    gradient3 = nll.gradient()
    assert frozenset(g.numpy() for g in gradient3) == pytest.approx(frozenset(both_gradients_true))


def test_simple_loss():
    true_a = 1.
    true_b = 4.
    true_c = -0.3
    truevals = true_a, true_b, true_c
    a_param = zfit.Parameter("variable_a15151loss", 1.5, -1., 20.,
                             step_size=z.constant(0.1))
    b_param = zfit.Parameter("variable_b15151loss", 3.5)
    c_param = zfit.Parameter("variable_c15151loss", -0.23)
    param_list = [a_param, b_param, c_param]

    def loss_func(params):
        a_param, b_param, c_param = params
        probs = z.convert_to_tensor((a_param - true_a) ** 2
                                    + (b_param - true_b) ** 2
                                    + (c_param - true_c) ** 4) + 0.42
        return tf.reduce_sum(input_tensor=tf.math.log(probs))

    with pytest.raises(ValueError):
        _ = zfit.loss.SimpleLoss(func=loss_func, params=param_list)

    loss_func.errordef = 1
    loss_deps = zfit.loss.SimpleLoss(func=loss_func, params=param_list)
    # loss = zfit.loss.SimpleLoss(func=loss_func)
    loss = zfit.loss.SimpleLoss(func=loss_func, params=param_list)
    loss2 = zfit.loss.SimpleLoss(func=loss_func, params=truevals)

    assert loss_deps.get_cache_deps() == set(param_list)
    assert set(loss_deps.get_params()) == set(param_list)

    loss_tensor = loss_func(param_list)
    loss_value_np = loss_tensor.numpy()

    assert loss.value().numpy() == pytest.approx(loss_value_np)
    assert loss_deps.value().numpy() == pytest.approx(loss_value_np)

    with pytest.raises(IntentionAmbiguousError):
        _ = loss + loss_deps

    minimizer = zfit.minimize.Minuit()
    result = minimizer.minimize(loss=loss)
    assert result.valid
    assert true_a == pytest.approx(result.params[a_param]['value'], rel=0.03)
    assert true_b == pytest.approx(result.params[b_param]['value'], rel=0.06)
    assert true_c == pytest.approx(result.params[c_param]['value'], rel=0.5)

    zfit.param.set_values(param_list, np.array(zfit.run(param_list)) + 0.6)
    result2 = minimizer.minimize(loss=loss2)
    assert result2.valid
    params = list(result2.params)
    assert true_a == pytest.approx(result2.params[params[0]]['value'], rel=0.03)
    assert true_b == pytest.approx(result2.params[params[1]]['value'], rel=0.06)
    assert true_c == pytest.approx(result2.params[params[2]]['value'], rel=0.5)


def test_create_new_nll():
    gaussian1, mu1, sigma1 = create_gauss1()
    gaussian2, mu2, sigma2 = create_gauss2()

    test_values = tf.constant(test_values_np)
    test_values = zfit.Data.from_tensor(obs=obs1, tensor=test_values)
    nll = zfit.loss.UnbinnedNLL(model=gaussian1, data=test_values)

    nll2 = nll.create_new(model=gaussian2)
    assert nll2.data[0] is nll.data[0]
    assert nll2.constraints == nll.constraints
    assert nll2._options == nll._options

    nll3 = nll.create_new()
    assert nll3.data[0] is nll.data[0]
    assert nll3.constraints == nll.constraints
    assert nll3._options == nll._options

    nll4 = nll.create_new(options={})
    assert nll4.data[0] is nll.data[0]
    assert nll4.constraints == nll.constraints
    assert nll4._options != nll._constraints


def test_create_new_extnll():
    gaussian1, mu1, sigma1, yield1 = create_gauss3ext()

    test_values = tf.constant(test_values_np)
    test_values = zfit.Data.from_tensor(obs=obs1, tensor=test_values)
    nll = zfit.loss.ExtendedUnbinnedNLL(model=gaussian1, data=test_values,
                                        constraints=zfit.constraint.GaussianConstraint(mu1, 1., 0.1))

    nll2 = nll.create_new(model=gaussian1)
    assert nll2.data[0] is nll.data[0]
    assert nll2.constraints == nll.constraints
    assert nll2._options == nll._options

    nll3 = nll.create_new()
    assert nll3.data[0] is nll.data[0]
    assert nll3.constraints == nll.constraints
    assert nll3._options == nll._options

    nll4 = nll.create_new(options={})
    assert nll4.data[0] is nll.data[0]
    assert nll4.constraints == nll.constraints
    assert nll4._options != nll._constraints


def test_create_new_simple():
    _, mu1, sigma1 = create_gauss1()

    loss = zfit.loss.SimpleLoss(lambda x, y: x * y,
                                params=[mu1, sigma1],
                                errordef=0.5)

    loss1 = loss.create_new()
    assert loss1._simple_func is loss._simple_func
    assert loss1.errordef == loss.errordef


@pytest.mark.parametrize('create_loss', [lambda: create_simultaneous_loss()[2], create_simpel_loss])
def test_callable_loss(create_loss):
    loss = create_loss()

    params = list(loss.get_params())
    x = np.array(zfit.run(params)) + 0.1
    value_loss = loss(x)
    with zfit.param.set_values(params, x):
        true_val = zfit.run(loss.value())
        assert true_val == pytest.approx(zfit.run(value_loss))
        with pytest.raises(BehaviorUnderDiscussion):
            assert true_val == pytest.approx(zfit.run(loss()))

    with pytest.raises(ValueError):
        loss(x[:-1])
    with pytest.raises(ValueError):
        loss(list(x) + [1])


@pytest.mark.parametrize('create_loss', [lambda: create_simultaneous_loss()[2], create_simpel_loss])
def test_iminuit_compatibility(create_loss):
    loss = create_loss()

    params = list(loss.get_params())
    x = np.array(zfit.run(params)) + 0.1
    zfit.param.set_values(params, x)

    with pytest.raises(ValueError):
        loss(x[:-1])
    with pytest.raises(ValueError):
        loss(list(x) + [1])

    import iminuit
    minimizer = iminuit.Minuit(loss, x)
    result = minimizer.migrad()
    assert result.valid
    minimizer.hesse()

    zfit.param.set_values(params, x)
    minimizer_zfit = zfit.minimize.Minuit()
    result_zfit = minimizer_zfit.minimize(loss)
    assert result_zfit.fmin == pytest.approx(result.fmin.fval, abs=0.03)



#  Copyright (c) 2021 zfit
import math

import pytest

import zfit


@pytest.mark.flaky(2)
def test_fail_on_nan_strategy():
    from zfit.minimizers.strategy import ToyStrategyFail
    sigma = zfit.Parameter("sigma", 2.)
    obs = zfit.Space("obs1", limits=(-4, 5))
    gauss = zfit.pdf.Gauss(1., sigma, obs=obs)

    sampler = gauss.create_sampler(3000)
    sampler.set_data_range(obs)
    nll = zfit.loss.UnbinnedNLL(model=gauss, data=sampler)
    minimizer = zfit.minimize.Minuit(strategy=ToyStrategyFail)
    sampler.resample()
    sigma.set_value(2.1)
    fitresult1 = minimizer.minimize(nll)
    assert fitresult1.converged

    sampler.resample()
    sigma.set_value(math.inf)
    fitresult2 = minimizer.minimize(nll)
    assert not fitresult2.converged
    assert fitresult2.edm == -999
    assert fitresult2.fmin == -999


def minimizers():
    return [
        zfit.minimize.Adam,
        zfit.minimize.NLoptMMAV1,
        zfit.minimize.IpyoptV1,
        zfit.minimize.Minuit,
        zfit.minimize.ScipySLSQPV1,
    ]


# sort for xdist: https://github.com/pytest-dev/pytest-xdist/issues/432
minimizers = sorted(minimizers(), key=lambda val: repr(val))


@pytest.mark.parametrize('minimizer_cls', minimizers)
def test_callback(minimizer_cls):
    from zfit.minimizers.strategy import PushbackStrategy

    class MyError(Exception):
        pass

    class MyStrategy(PushbackStrategy):
        def callback(*args, **kwargs):
            raise MyError

    minimizer = zfit.minimize.ScipySLSQPV1(strategy=MyStrategy)
    loss = lambda *args, **kwargs: 42.
    loss.errordef = 0.5
    with pytest.raises(MyError):
        minimizer.minimize(loss, [1, 3])



#  Copyright (c) 2021 zfit
import numpy as np
import pytest
import scipy
import scipy.special
import tensorflow as tf

import zfit


def func_cosxy2_np(x, y):
    return np.cos(x * y ** 2)


def func_cosxy2_tf(x, y):
    return tf.cos(x * y ** 2)


def integral_y(x, lowery, uppery):
    def indef_limit(limit):
        integral = (np.sqrt(2) * np.sqrt(np.pi) * scipy.special.fresnel(
            np.sqrt(2) * limit * np.sqrt(x) / np.sqrt(np.pi))[1] * scipy.special.gamma(1 / 4) / (
                        8 * np.sqrt(x) * scipy.special.gamma(5 / 4)))
        return integral

    return indef_limit(uppery) - indef_limit(lowery)


def integral_x(y, lowerx, upperx):
    def indef_limit(limit):
        return np.sin(limit * y ** 2) / y ** 2

    return indef_limit(upperx) - indef_limit(lowerx)


def test_partial_integral():
    class CosXY2(zfit.pdf.ZPDF):
        _PARAMS = []

        def _unnormalized_pdf(self, x):
            x, y = x.unstack_x()
            return func_cosxy2_tf(x=x, y=y)

    xspace = zfit.Space("x", limits=(1, 4))
    yspace = zfit.Space("y", limits=(1, 4))
    obs = xspace * yspace

    cosxy2 = CosXY2(obs=obs)

    data_np = np.linspace((1, 1), (4, 4), 4000)
    data = zfit.Data.from_numpy(obs=obs, array=data_np)
    probs = cosxy2.pdf(data)
    probs_np = probs.numpy()
    x = data_np[:, 0]
    y = data_np[:, 1]
    probs_func = func_cosxy2_np(x=x, y=y)
    ratios = probs_np / probs_func
    assert pytest.approx(0., rel=0.0001) == np.std(ratios)  # ratio should be constant
    ratio = np.average(ratios)

    integral_x_tf = cosxy2.partial_integrate(x=data, limits=xspace)
    integral_y_tf = cosxy2.partial_integrate(x=data, limits=yspace)
    integral_x_np = integral_x_tf.numpy()
    integral_y_np = integral_y_tf.numpy()

    lowerx, upperx = xspace.rect_limits
    lowery, uppery = yspace.rect_limits
    integral_x_true = integral_x(y=y, lowerx=lowerx[0], upperx=upperx[0]) * ratio
    integral_y_true = integral_y(x=x, lowery=lowery[0], uppery=uppery[0]) * ratio

    np.testing.assert_allclose(integral_x_true, integral_x_np, atol=3.5e-3)
    np.testing.assert_allclose(integral_y_true, integral_y_np, atol=3.5e-3)



#  Copyright (c) 2021 zfit
from functools import wraps
from typing import Any, Iterable, Tuple, Union

import tensorflow as tf
import tensorflow_probability as tfp

from .zextension import function as function

__all__ = ["counts_multinomial", "sample_with_replacement"]

from ..settings import ztypes


def sample_with_replacement(a: tf.Tensor, axis: int, sample_shape: Tuple[int]) -> tf.Tensor:
    """Sample from `a` with replacement to return a Tensor with `sample_shape`.

    Args:
        a (): Tensor to sample from
        axis (): int axis to sample along
        sample_shape (): Shape of the new samples along the axis

    Returns:
        tf.Tensor of shape a.shape[:axis] + samples_shape + a.shape[axis + 1:]


    Examples:

    .. code:: pycon

        >>> a = tf.random.uniform(shape=(10, 20, 30), dtype=tf.float32)
        >>> random_choice(a, axis=0)
        <tf.Tensor 'GatherV2:0' shape=(1, 20, 30) dtype=float32>
        >>> random_choice(a, axis=1, samples_shape=(2, 3))
        <tf.Tensor 'GatherV2_2:0' shape=(10, 2, 3, 30) dtype=float32
        >>> random_choice(a, axis=0, samples_shape=(100,))
        <tf.Tensor 'GatherV2_3:0' shape=(100, 20, 30) dtype=float32>
    """

    dim = tf.shape(a)[axis]
    choice_indices = tf.random.uniform(sample_shape, minval=0, maxval=dim, dtype=tf.int32)
    samples = tf.gather(a, choice_indices, axis=axis)
    return samples


def counts_multinomial(total_count: Union[int, tf.Tensor], probs: Iterable[Union[float, tf.Tensor]] = None,
                       logits: Iterable[Union[float, tf.Tensor]] = None, dtype=tf.int64) -> tf.Tensor:
    """Get the number of counts for different classes with given probs/logits.

    Args:
        total_count: The total number of draws.
        probs: Length k (number of classes) object where the k-1th entry contains the probability to
            get a single draw from the class k. Have to be from [0, 1] and sum up to 1.
        logits: Same as probs but from [-inf, inf] (will be transformet to [0, 1])

    Returns:
        Shape (k,) tensor containing the number of draws.
    """
    from .. import z

    total_count = tf.convert_to_tensor(total_count)
    probs = z.convert_to_tensor(probs) if probs is not None else probs
    logits = tf.convert_to_tensor(logits) if logits is not None else logits

    if probs is not None:
        probs = tf.cast(probs, dtype=tf.float64)
        float_dtype = probs.dtype
    elif logits is not None:
        logits = tf.cast(logits, tf.float64)
        float_dtype = logits.dtype
    else:
        raise ValueError("Exactly one of `probs` or`logits` have to be specified")
    total_count = tf.cast(total_count, dtype=float_dtype)

    # needed since otherwise shape of sample will be (1, n_probs)
    # total_count = tf.broadcast_to(total_count, shape=probs_logits_shape)

    @function
    def wrapped_func(dtype, logits, probs, total_count):

        dist = tfp.distributions.Multinomial(total_count=total_count, probs=probs, logits=logits)
        counts = dist.sample()
        counts = tf.cast(counts, dtype=dtype)
        return counts

    return wrapped_func(dtype, logits, probs, total_count)


@wraps(tf.random.normal)
def normal(shape, mean=0.0, stddev=1.0, dtype=ztypes.float, seed=None, name=None):
    return tf.random.normal(shape=shape, mean=mean, stddev=stddev, dtype=dtype, seed=seed, name=name)


@wraps(tf.random.uniform)
def uniform(shape, minval=0, maxval=None, dtype=ztypes.float, seed=None, name=None):
    return tf.random.uniform(shape=shape, minval=minval, maxval=maxval, dtype=dtype, seed=seed, name=name)


@wraps(tf.random.poisson)
def poisson(lam: Any, shape: Any, dtype: tf.DType = ztypes.float, seed: Any = None, name: Any = None):
    return tf.random.poisson(lam=lam, shape=shape, dtype=dtype, seed=seed, name=name)



#  Copyright (c) 2021 zfit

import tensorflow as tf

import zfit
from zfit import z

# IMPORTANT! The communication of which axis corresponds to which data point happens here. So the user knows now that
# he should create this pdf with a space in the obs (x, y, z).


class CustomPDF(zfit.pdf.ZPDF):
    """3-dimensional PDF calculating doing something fancy, takes x, y, z as data."""
    _PARAMS = ['alpha', 'beta']  # specify which parameters to take
    _N_OBS = 3

    def _unnormalized_pdf(self, x):  # implement function
        x, y, z = x.unstack_x()
        alpha = self.params['alpha']
        beta = self.params['beta']
        x_new = tf.math.cos(alpha) * x
        y_new = tf.math.sinh(beta) * y
        z_new = z + 4.2
        return x_new ** 2 + y_new ** 2 + z_new ** 2


xobs = zfit.Space('xobs', (-4, 4))
yobs = zfit.Space('yobs', (-3, 3))
zobs = zfit.Space('z', (-2, 2))
obs = xobs * yobs * zobs

alpha = zfit.Parameter("alpha", 0.2)  # floating
beta = zfit.Parameter("beta", 0.4, floating=False)  # non-floating
custom_pdf = CustomPDF(obs=obs, alpha=alpha, beta=beta)

integral = custom_pdf.integrate(limits=obs)  # = 1 since normalized
sample = custom_pdf.sample(n=1000)  # DO NOT USE THIS FOR TOYS!
prob = custom_pdf.pdf(sample)  # DO NOT USE THIS FOR TOYS!

integral_np, sample_np, prob_np = [integral.numpy(), sample.numpy(), prob.numpy()]



#  Copyright (c) 2021 zfit

from .core.parameter import (ComplexParameter, ComposedParameter,
                             ConstantParameter, Parameter, assign_values,
                             convert_to_parameter, convert_to_parameters,
                             set_values)

__all__ = ['ConstantParameter', 'Parameter', 'ComposedParameter', 'ComplexParameter', 'convert_to_parameter',
           'convert_to_parameters',
           'set_values', 'assign_values']



#  Copyright (c) 2021 zfit
import numpy as np
import pytest
import scipy.stats

import zfit
from zfit import z
from zfit.z.math import (autodiff_gradient, autodiff_hessian,
                         numerical_gradient, numerical_hessian)


def test_numerical_gradient():
    param1 = zfit.Parameter('param1', 4.)
    param2 = zfit.Parameter('param2', 5.)
    param3 = zfit.Parameter('param3', 2.)

    def func1():
        return param1 * param2 ** 2 + param3 ** param1

    num_gradients = numerical_gradient(func1, params=[param1, param2, param3])
    tf_gradients = autodiff_gradient(func1, params=[param1, param2, param3])
    np.testing.assert_allclose(num_gradients, tf_gradients)


@pytest.mark.parametrize('graph', [False, True])
def test_numerical_hessian(graph):
    param1 = zfit.Parameter('param1', 4.)
    param2 = zfit.Parameter('param2', 5.)
    param3 = zfit.Parameter('param3', 2.)

    def func1():
        return param1 * param2 ** 2 + param3 ** param1

    def create_derivatives(func1, params):
        num_hessian_diag = numerical_hessian(func1, params=params, hessian='diag')
        num_hessian = numerical_hessian(func1, params=params)
        tf_hessian_diag = autodiff_hessian(func1, params=params, hessian='diag')
        tf_hessian = autodiff_hessian(func1, params=params)
        return num_hessian, num_hessian_diag, tf_hessian, tf_hessian_diag

    params = [param1, param2, param3]
    if graph:
        create_derivatives = z.function(create_derivatives)
    num_hessian, num_hessian_diag, tf_hessian, tf_hessian_diag = create_derivatives(func1, params)
    np.testing.assert_allclose(num_hessian, tf_hessian, rtol=1e-4, atol=1e-10)
    tf_hessian_diag_from_hessian = [tf_hessian[i, i] for i in range(len(params))]
    np.testing.assert_allclose(tf_hessian_diag_from_hessian, tf_hessian_diag, rtol=1e-4, atol=1e-10)
    np.testing.assert_allclose(num_hessian_diag, tf_hessian_diag, rtol=1e-4, atol=1e-10)


def test_reduce_geometric_mean():
    rnd1 = np.random.poisson(1000, size=(54, 14, 3)).astype(np.float64)
    gmean_np = scipy.stats.mstats.gmean(rnd1, axis=None)
    gmea_z = z.math.reduce_geometric_mean(rnd1)
    np.testing.assert_allclose(gmea_z, gmean_np)



"""Module for caching.

The basic concept of caching in Zfit builds on a "cacher", that caches a certain value and that
is dependent of "cache_dependents". By implementing `ZfitCachable`, an object will be able to play both
roles. And most importantly, it has a `_cache` dict, that contains all the cache.

Basic principle
===============

A "cacher" adds any dependents that it may comes across with `add_cache_dependents`. For example,
for a loss this would be all pdfs and data. Since :py:class:`~zfit.Space` is immutable, there is no need to add this
as a dependent. This leads to the "cache_dependent" to register the "cacher" and to remember it.

In case, any "cache_dependent" changes in a way the cache of itself (and any "cacher") is invalid,
which is done in the simplest case by decorating a method with `@invalidates_cache`, the "cache_dependent":

 * clears it's own cache with `reset_cache_self` and
 * "clears" any "cacher"s cache with `reset_cache(reseter=self)`, telling the "cacher" that it should
   reset the cache. This is also where more fine-grained control (depending on which "cache_dependent"
   calls `reset_cache`) can be brought into play.

Example with a pdf that caches the normalization:

.. code:: python

    class Parameter(Cachable):
        def load(new_value):  # does not require to build a new graph
            # do something

        @invalidates_cache
        def change_limits(new_limits):  # requires to build a new graph (as an example)
            # do something

    # create param1, param2 from `Parameter`

    class MyPDF(Cachable):
        def __init__(self, param1, param2):
            self.add_cache_dependents([param1, param2])

        def cached_func(...):
            if self._cache.get('my_name') is None:
                result = ...  # calculations here
                self._cache['my_name']
            else:
                result = self._cache['my_name']
            return result
"""

#  Copyright (c) 2021 zfit

import functools
import weakref
from abc import abstractmethod
from typing import Iterable, Mapping, Union

import numpy as np
import tensorflow as tf

from . import ztyping
from .container import convert_to_container


class ZfitGraphCachable:

    @abstractmethod
    def register_cacher(self, cacher: "ZfitGraphCachable"):
        raise NotImplementedError

    @abstractmethod
    def add_cache_deps(self, cache_dependents, allow_non_cachable):
        """Add dependents that render the cache invalid if they change.

        Args:
            cache_dependents:
            allow_non_cachable: If `True`, allow `cache_dependents` to be non-cachables.
                If `False`, any `cache_dependents` that is not a `ZfitCachable` will raise an error.

        Raises:
            TypeError: if one of the `cache_dependents` is not a `ZfitCachable` _and_ `allow_non_cachable`
                if `False`.
        """
        pass

    @abstractmethod
    def reset_cache_self(self):
        """Clear the cache of self and all dependent cachers."""
        pass

    @abstractmethod
    def reset_cache(self, reseter):
        pass


class GraphCachable(ZfitGraphCachable):
    graph_caching_methods = []
    instances = weakref.WeakSet()

    def __init__(self, *args, **kwargs):
        self._cache = {}
        self._cachers = weakref.WeakKeyDictionary()
        self.reset_cache_self()
        self.instances.add(self)
        super().__init__(*args, **kwargs)

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        graph_caching_methods = []
        for func_name in dir(cls):
            if not func_name.startswith("__"):
                func = getattr(cls, func_name)
                if callable(func) and hasattr(func, 'zfit_graph_cache_registered'):
                    # assert hasattr(func, "_descriptor_cache"), "TensorFlow internals have changed. Need to update cache"
                    func.zfit_graph_cache_registered = True
                    graph_caching_methods.append(func)
        cls.graph_caching_methods = graph_caching_methods

    def register_cacher(self, cacher: ztyping.CacherOrCachersType):
        """Register a `cacher` that caches values produces by this instance; a dependent.

        Args:
            cacher:
        """
        if not isinstance(cacher, ZfitGraphCachable):
            raise TypeError(f"`cacher` is not a `ZfitCachable` but {type(cacher)}")
        if not cacher in self._cachers:
            self._cachers[cacher] = None  # could we have a more useful value?

    def add_cache_deps(self, cache_deps: ztyping.CacherOrCachersType, allow_non_cachable: bool = True):
        """Add dependencies that render the cache invalid if they change.

        Args:
            cache_deps:
            allow_non_cachable: If `True`, allow `cache_dependents` to be non-cachables.
                If `False`, any `cache_dependents` that is not a `ZfitCachable` will raise an error.

        Raises:
            TypeError: if one of the `cache_dependents` is not a `ZfitCachable` _and_ `allow_non_cachable`
                if `False`.
        """
        cache_deps = convert_to_container(cache_deps)
        for cache_dep in cache_deps:
            if isinstance(cache_dep, ZfitGraphCachable):
                cache_dep.register_cacher(self)
            elif not allow_non_cachable:
                raise TypeError("cache_dependent {} is not a `ZfitCachable` but {}".format(cache_dep,
                                                                                           type(cache_dep)))

    def reset_cache_self(self):
        """Clear the cache of self and all dependent cachers."""
        self._clean_cache()
        self._inform_cachers()

    def reset_cache(self, reseter: 'ZfitGraphCachable'):
        self.reset_cache_self()

    def _clean_cache(self):
        # for func_holder in self.graph_caching_methods:
        #     func_holder.reset
        self._cache = {}
        return

    def _inform_cachers(self):
        for cacher in self._cachers:
            cacher.reset_cache(reseter=self)


def invalidate_graph(func):
    @functools.wraps(func)
    def wrapped_func(*args, **kwargs):
        self = args[0]
        if not isinstance(self, ZfitGraphCachable):
            raise TypeError("Decorator can only be used in a subclass of `ZfitCachable`")
        self.reset_cache(reseter=self)

        return func(*args, **kwargs)

    return wrapped_func


class FunctionCacheHolder(GraphCachable):
    IS_TENSOR = object()

    def __init__(self, func, wrapped_func,
                 cachables: Union[ZfitGraphCachable, object, Iterable[Union[ZfitGraphCachable, object]]] = None,
                 cachables_mapping=None):
        """`tf.function` decorated function holder with caching dependencies on inputs.

        A `tf.function` creates a new graph for every signature that is encountered. It automatically caches them but
        thereby assumes that Python objects are immutable. Any mutation won't be detected. Therefore, an extra wrapper
        is needed. The input signature is compared with firstly checking whether the function is the same and then
        doing an equal comparison of the arguments (maybe too costly?).

        The `FunctionCacheHolder` holds the
         - original python function which serves as the hash of the object
         - wrapped python function, `wrapped_func`
         - the (keyword-)arguments

        If any of the keyword arguments changes in a way that the graph cache is invalid, this holder will have
        `is_valid` set to False and the `wrapped_func` cannot be used anymore, instead a new `tf.function` should
        be created as a call to the `wrapped_func` with the given arguments will result in an outdated graph.

        Args:
            func: Python function that serves as a hash of the holder. Notice that equality is different
                defined.
            wrapped_func: Wrapped `func` with `tf.function`. The holder signals via
                `is_valid` whether this function is still valid to be used.
            cachables: objects that are cached. If they change, the cache is invalidated
            cachables_mapping: keyword arguments to the function. If the values change, the cache is
                invalidated.
        """
        # cache = {} if cache is None else cache
        self.delete_from_cache = False
        self.wrapped_func = wrapped_func
        # self.parent_cache = cache
        self.python_func = func
        self._hash_value = hash(self.python_func)
        if cachables is None and cachables_mapping is None:
            raise ValueError("Both `cachables and `cachables_mapping` are None. One needs to be different from None.")
        if cachables is None:
            cachables = []
        if cachables_mapping is None:
            cachables_mapping = {}
        cachables = convert_to_container(cachables, container=list)
        cachables_values = convert_to_container(cachables_mapping.values(), container=list)
        cachables_all = cachables + cachables_values
        self.immutable_representation = self.create_immutable(cachables, cachables_mapping)
        # self._hash_value = hash(self.immutable_representation)
        super().__init__()  # resets the cache
        self.add_cache_deps(cachables_all)
        self.is_valid = True  # needed to make the cache valid again

    def reset_cache_self(self):
        self.is_valid = False

    def create_immutable(self, args, kwargs):
        """Create a tuple of the args and kwargs by combining them as args + kwargs.keys() + kwargs.values()`

        Args:
            args: list like
            kwargs: dict-like

        Returns:
        """
        # is initialized before the core
        from ..core.interfaces import ZfitData, ZfitParameter, ZfitSpace

        args = list(args)
        kwargs = list(kwargs.keys()) + list(kwargs.values())
        combined = []
        if args != []:
            combined += args
        if kwargs != []:
            combined += args
        combined_cleaned = []
        for obj in combined:
            if isinstance(obj, ZfitData):
                obj = (id(obj),)

            elif isinstance(obj, ZfitParameter):
                obj = (ZfitParameter, obj.name)
            elif isinstance(obj, ZfitSpace):
                obj = (id(obj),)
            elif tf.is_tensor(obj):
                obj = self.IS_TENSOR
            elif isinstance(obj, np.ndarray):
                obj = (obj,) if sum(obj.shape) < 20 else id(obj)
            combined_cleaned.append(obj)

        return tuple(combined_cleaned)

    def __hash__(self) -> int:
        return self._hash_value

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FunctionCacheHolder):
            return False
        # return all(obj1 == obj2 for obj1, obj2 in zip(self.immutable_representation, other.immutable_representation))
        array_repr_self = np.array(self.immutable_representation, dtype=object)
        array_repr_other = np.array(other.immutable_representation, dtype=object)
        try:
            return all(np.equal(array_repr_self, array_repr_other))
        except ValueError:  # broadcasting does not work
            return False
        except TypeError:  # OperatorNotAllowedError inherits from this
            return False
        # TODO: activate the below? costly, but runs?
        # except OperatorNotAllowedInGraphError:  # we have to assume they're not the same
        #     return False

    def __repr__(self) -> str:
        return f"<FunctionCacheHolder: {self.python_func}, valid={self.is_valid}>"


def clear_graph_cache():
    from zfit.z.zextension import FunctionWrapperRegistry

    for registry in FunctionWrapperRegistry.registries:
        for all_meth in registry.function_cache.values():
            for wrapped_meth in all_meth:
                wrapped_meth = wrapped_meth.wrapped_func
                wrapped_meth._created_variables = None
                wrapped_meth._stateful_fn = None
                wrapped_meth._stateless_fn = None
                wrapped_meth._descriptor_cache.clear()

    for registry in FunctionWrapperRegistry.registries:
        registry.reset()
    for instance in GraphCachable.instances:
        instance.reset_cache('global')
    # Cachable.graph_caching_methods.clear()
    tf.compat.v1.reset_default_graph()
